import { Direction, EntityItemComponent, world, EntityEquippableComponent, EquipmentSlot, system, Player, GameMode, ItemStack, EntityVariantComponent, EntityInventoryComponent, EntityRideableComponent, EntityRidingComponent, TicksPerSecond } from '@minecraft/server';
import { ModalFormData, ActionFormData } from '@minecraft/server-ui';

/**
 * Calculates the distance between two vectors
 * @param vectorA
 * @param vectorB
 * @returns Distance as number
 */
function getDistance(vectorA, vectorB) {
    const x = vectorA.x - vectorB.x;
    const y = vectorA.y - vectorB.y;
    const z = vectorA.z - vectorB.z;
    return Math.sqrt(x * x + y * y + z * z);
}
/**
 * Adds one vector to another
 * @param vectorA
 * @param vectorB
 * @returns Result as Vector
 */
function addVector(vectorA, vectorB) {
    return {
        x: vectorA.x + (vectorB.x || 0),
        y: vectorA.y + (vectorB.y || 0),
        z: vectorA.z + (vectorB.z || 0),
    };
}
function vectorsEqual(a, b) {
    return a.x === b?.x && a.y === b?.y && a.z === b?.z;
}

// NORTH must be 0, aka default rotation
const allDirections = ["North", "NorthEast", "East", "SouthEast", "South", "SouthWest", "West", "NorthWest"];
const cardinalDirections = ["North", "East", "South", "West"];
function getCardinal(permutation) {
    const cardinal = permutation?.getState("minecraft:cardinal_direction");
    if (!cardinal)
        return null;
    const direction = `${cardinal[0].toUpperCase()}${cardinal.slice(1)}`;
    if (direction === Direction.Up || direction === Direction.Down)
        return null;
    return direction;
}
function getRailDirections(location, permutation, checkNeighbours = true, hailingFrom) {
    if (!permutation?.hasTag("pod_trn:rail") || !location)
        return null;
    const details = getRailDetails(permutation);
    const cardinal = details.cardinal;
    let front = null;
    let side = null;
    let back = null;
    let frontY = 0;
    let backY = 0;
    switch (details.type) {
        case "conventional": {
            const { bend, railDirection, slope } = details;
            if (bend && railDirection === "straight")
                throw new Error("Disallowed state on rail");
            if (slope === 4)
                frontY = -1;
            if (slope === 3)
                backY = 1;
            front = "North";
            back = "South";
            switch (railDirection) {
                case "left":
                    front = "NorthWest";
                    if (!bend)
                        back = "SouthEast";
                    break;
                case "right":
                    front = "NorthEast";
                    if (!bend)
                        back = "SouthWest";
                    break;
            }
            break;
        }
        case "switch": {
            const { switchedDirection } = details;
            side = switchedDirection === "left" ? "NorthWest" : "NorthEast";
            front = "North";
            back = "South";
            break;
        }
        case "buffer":
            const { railDirection } = details;
            switch (railDirection) {
                case "left":
                    back = "SouthEast";
                    break;
                case "right":
                    back = "SouthWest";
                    break;
                default:
                    back = "South";
            }
            front = null;
            break;
    }
    front = directionRelativeTo(front, cardinal);
    back = directionRelativeTo(back, cardinal);
    side = directionRelativeTo(side, cardinal);
    if (hailingFrom) {
        const hailingFromReciprocal = getReciprocal(hailingFrom);
        if (hailingFromReciprocal === front)
            front = null;
        if (hailingFromReciprocal === back)
            back = null;
        if (hailingFromReciprocal === side)
            side = null;
    }
    const connectedTo = [];
    if (front)
        connectedTo.push(front);
    if (back)
        connectedTo.push(back);
    if (side)
        connectedTo.push(side);
    const resultObject = {
        East: false,
        North: false,
        NorthEast: false,
        NorthWest: false,
        South: false,
        SouthEast: false,
        SouthWest: false,
        West: false,
        details,
        connectedTo
    };
    if (!checkNeighbours) {
        resultObject.checked = false;
        if (front) {
            resultObject[front] = true;
        }
        if (back) {
            resultObject[back] = true;
        }
        if (side) {
            resultObject[side] = true;
        }
        return resultObject;
    }
    resultObject.checked = true;
    if (front) {
        const frontBlock = getNeighbour(location, front, frontY);
        const frontNeighbour = getRailDirections(frontBlock?.block, frontBlock?.block?.permutation, false);
        resultObject[front] = {
            hasConnectedNeighbour: !!frontNeighbour?.[getReciprocal(front)],
            step: backY !== 0,
            block: frontBlock?.block ?? null,
            traversable: details.type === "switch" ? !details.switched : true
        };
    }
    if (back) {
        const backBlock = getNeighbour(location, back, backY);
        const backNeighbour = getRailDirections(backBlock?.block, backBlock?.block?.permutation, false);
        resultObject[back] = {
            hasConnectedNeighbour: !!backNeighbour?.[getReciprocal(back)],
            step: backY !== 0,
            block: backBlock?.block ?? null,
            traversable: true // Always traversable because of currently added rails, may need to change eventually
        };
    }
    if (side) {
        const sideBlock = getNeighbour(location, side, backY);
        const sideNeighbour = getRailDirections(sideBlock?.block, sideBlock?.block?.permutation, false);
        resultObject[side] = {
            hasConnectedNeighbour: !!sideNeighbour?.[getReciprocal(side)],
            step: false,
            block: sideBlock?.block ?? null,
            traversable: details.type === "switch" ? details.switched : true
        };
    }
    return resultObject;
}
function getReciprocal(direction) {
    for (const [key, value] of Object.entries({
        "NorthWest": "SouthEast",
        "North": "South",
        "NorthEast": "SouthWest",
        "East": "West",
    })) {
        if (direction === key)
            return value;
        if (direction === value)
            return key;
    }
    return "North";
}
function directionRelativeTo(direction, local) {
    if (!direction)
        return null;
    const use = allDirections;
    const offset = use.findIndex(d => d === local);
    const global = use.findIndex(d => d === direction);
    return use[(global + offset) % use.length];
}
function getNeighbour(location, direction, dy = 0) {
    let x = 0, z = 0;
    if (direction.includes("North"))
        z = -1;
    else if (direction.includes("South"))
        z = 1;
    if (direction.includes("West"))
        x = -1;
    else if (direction.includes("East"))
        x = 1;
    const block = location.dimension.getBlock(addVector(location, {
        x,
        z,
        y: dy
    }));
    if (!block?.isValid())
        return null;
    const details = getRailDetails(block.permutation);
    if (!details)
        return null;
    return {
        block,
        ...details
    };
}
function conventionalDirectionToRailCardinal(direction) {
    if (cardinalDirections.includes(direction)) {
        const cardinal = direction;
        return { railDirection: "straight", cardinal };
    }
    const first = direction.slice(0, 5);
    const second = direction.slice(5);
    if (first === "South")
        return {
            cardinal: "South",
            railDirection: second === "East" ? "left" : "right"
        };
    return {
        cardinal: "North",
        railDirection: second === "East" ? "right" : "left"
    };
}
function switchDirectionToSwitchCardinal(direction, primarySideways = false) {
    if (cardinalDirections.includes(direction)) {
        const cardinal = direction;
        return { cardinal };
    }
    const first = direction.slice(0, 5);
    const second = direction.slice(5);
    if (primarySideways) {
        if (second === "East")
            return {
                cardinal: "East",
                switchedDirection: first === "North" ? "left" : "right"
            };
        return {
            cardinal: "West",
            switchedDirection: first === "North" ? "right" : "left"
        };
    }
    if (first === "South")
        return {
            cardinal: "South",
            switchedDirection: second === "East" ? "left" : "right"
        };
    return {
        cardinal: "North",
        switchedDirection: second === "East" ? "right" : "left"
    };
}
/**
 * Make a rail point towards a target
 *
 *
 * @param anchor
 * @param target
 * @param currentDetails
 */
function rotate(anchor, target, currentDetails) {
    // This function needs to return false if the requested configuration is illegal (state-wise)
    const targetReciprocal = getReciprocal(target);
    const convertedTarget = conventionalDirectionToRailCardinal(targetReciprocal);
    switch (currentDetails.type) {
        case "conventional": {
            if (anchor.length > 1)
                return false;
            if (!anchor.length)
                return { type: "conventional", ...convertedTarget };
            const firstAnchor = anchor.at(0);
            const separation = allDirections.findIndex(d => d === firstAnchor) - allDirections.findIndex(d => d === targetReciprocal);
            const targetCorner = target.length > 5;
            let curveDirection = targetCorner ? getReciprocal(firstAnchor) : targetReciprocal;
            switch (separation) {
                case 0:
                    return {
                        type: "conventional",
                        bend: false, ...convertedTarget
                    };
                case -7: // Essentially 1, due to being right at the edge of the direction array
                case 1:
                    return {
                        type: "conventional",
                        bend: true,
                        railDirection: targetCorner ? "left" : "right",
                        cardinal: curveDirection
                    };
                case -1:
                case 7: // Again, essentially -1, for same reasons
                    return {
                        type: "conventional",
                        bend: true,
                        railDirection: targetCorner ? "right" : "left",
                        cardinal: curveDirection
                    };
                default:
                    return false;
            }
        }
        case "switch": {
            if (anchor.length > 2)
                return false;
            const convertedTarget = switchDirectionToSwitchCardinal(target, anchor.some(d => d === "West" || d === "East"));
            if (!anchor.length) {
                return { type: "switch", ...convertedTarget };
            }
            const [firstAnchor, secondAnchor] = anchor;
            if (anchor.length === 1) {
                if (firstAnchor.length <= 5 && target.length > 5) {
                    return { type: "switch", ...convertedTarget };
                }
                if (firstAnchor.length > 5 && target.length <= 5) {
                    const convertedTarget = switchDirectionToSwitchCardinal(firstAnchor, target === "West" || target === "East");
                    return { type: "switch", ...convertedTarget };
                }
                return false;
            }
            if (target.length <= 5)
                return false;
            return { type: "switch", ...convertedTarget };
        }
        case "buffer":
            if (anchor.length)
                return false;
            return { ...currentDetails, ...convertedTarget };
        default:
            // The default config is only having one valid cardinal direction, that being north
            if (anchor.length)
                return false;
            if (!cardinalDirections.includes(target))
                return false;
            return {
                ...currentDetails,
                cardinal: targetReciprocal
            };
    }
}
function getRailDetails(permutation) {
    if (!permutation.hasTag("pod_trn:rail"))
        return null;
    const cardinal = getCardinal(permutation);
    if (permutation.hasTag("pod_trn:conventional_rail")) {
        return {
            type: "conventional",
            slope: permutation.getState("pod_trn:slope"),
            bend: permutation.getState("pod_trn:bend"),
            railDirection: permutation.getState("pod_trn:rail_direction"),
            cardinal
        };
    }
    if (permutation.hasTag("pod_trn:buffer_rail")) {
        return {
            type: "buffer",
            railDirection: permutation.getState("pod_trn:rail_direction"),
            cardinal
        };
    }
    if (permutation.hasTag("pod_trn:switched_rail")) {
        return {
            type: "switch",
            switched: permutation.getState("pod_trn:switched"),
            switchedDirection: permutation.getState("pod_trn:switched_direction"),
            cardinal
        };
    }
    return {
        type: "unknown",
        cardinal
    };
}
function applyRailDetails(permutation, details) {
    switch (details.type) {
        case "conventional":
            if (details.slope !== undefined)
                permutation = permutation.withState("pod_trn:slope", details.slope);
            if (details.bend !== undefined)
                permutation = permutation.withState("pod_trn:bend", details.bend);
        // Fall through
        case "buffer":
            if (details.railDirection !== undefined)
                permutation = permutation.withState("pod_trn:rail_direction", details.railDirection);
            break;
        case "switch":
            if (details.switched !== undefined)
                permutation = permutation.withState("pod_trn:switched", details.switched);
            if (details.switchedDirection !== undefined)
                permutation = permutation.withState("pod_trn:switched_direction", details.switchedDirection);
            break;
    }
    if (details.cardinal !== undefined)
        permutation = permutation.withState("minecraft:cardinal_direction", details.cardinal.toLowerCase());
    return permutation;
}
function slopedRailCheck(block, validDirections) {
    const railDirections = getRailDirections(block, block.permutation);
    if (!railDirections || !railDirections.checked)
        return "CONTINUE";
    const [front, back] = railDirections.connectedTo;
    return !(!(validDirections.includes(front) && validDirections.includes(back)) &&
        ((railDirections[front] && railDirections[front]?.hasConnectedNeighbour) ||
            (railDirections[back] && railDirections[back]?.hasConnectedNeighbour)));
}
function autoConnectRail(location, permutation) {
    // Cancel if rail below or not air above
    if (location.dimension.getBlock(addVector(location, { y: -1 }))?.hasTag("pod_trn:rail"))
        return false;
    if (!location.dimension.getBlock(addVector(location, { y: 1 }))?.matches("air"))
        return false;
    const ownDetails = getRailDetails(permutation);
    // Slope top handling
    if (ownDetails.type === "conventional")
        for (const direction of cardinalDirections) {
            const neighbourAbove = getNeighbour(location, direction, 1);
            const neighbourBelow = getNeighbour(location, direction, -1);
            const reciprocal = getReciprocal(direction);
            const both = [direction, reciprocal];
            if (neighbourAbove && neighbourAbove.type === "conventional") {
                const result = slopedRailCheck(neighbourAbove.block, both);
                if (result === "CONTINUE")
                    continue;
                if (!result)
                    return false;
                if (neighbourAbove.slope !== 0 && neighbourAbove.slope !== 1 && neighbourAbove.slope !== 4)
                    return false;
                if (neighbourAbove.slope <= 1) {
                    const above = location.dimension.getBlock(addVector(location, { y: 1 }));
                    if (!above?.isAir)
                        return false;
                    above.setType(permutation.type.id);
                    above.setPermutation(applyRailDetails(permutation, {
                        type: "conventional",
                        cardinal: reciprocal,
                        slope: 4
                    }));
                    location.dimension.playSound("dig.stone", location);
                }
                neighbourAbove.block.setPermutation(applyRailDetails(neighbourAbove.block.permutation, {
                    type: "conventional",
                    cardinal: reciprocal
                }));
                let neighbour = null;
                for (let i = neighbourAbove.slope <= 1 ? 3 : 2; i > 0; i--) {
                    const newNeighbour = getNeighbour(neighbour?.block ?? location, reciprocal);
                    if (!newNeighbour)
                        break;
                    neighbour = newNeighbour;
                    const result = slopedRailCheck(neighbour?.block, both);
                    if (result === "CONTINUE")
                        continue;
                    if (!result)
                        return false;
                    const slope = i;
                    neighbour.block.setPermutation(applyRailDetails(neighbour.block.permutation, {
                        type: "conventional",
                        cardinal: reciprocal,
                        slope
                    }));
                }
                return neighbourAbove.slope <= 1 ? false : applyRailDetails(permutation, {
                    type: "conventional",
                    cardinal: reciprocal,
                    slope: 3
                });
            }
            if (neighbourBelow && neighbourBelow.type === "conventional") {
                let neighbour = neighbourBelow;
                const batchChanges = [];
                for (let i = 4; i > 0; i--) {
                    const result = slopedRailCheck(neighbour.block, both);
                    if (result === "CONTINUE")
                        continue;
                    if (!result)
                        return false;
                    const currentNeighbour = neighbour;
                    if (i === 4) {
                        const above = currentNeighbour.block?.above();
                        if (!above?.isAir)
                            return false;
                        above.setType(currentNeighbour.block?.typeId);
                        above.setPermutation(applyRailDetails(currentNeighbour.block.permutation, {
                            type: "conventional",
                            slope: i,
                            cardinal: direction
                        }));
                        currentNeighbour.block?.setType("air");
                    }
                    else {
                        batchChanges.push(() => currentNeighbour.block.setPermutation(applyRailDetails(currentNeighbour.block.permutation, {
                            type: "conventional",
                            slope: i,
                            cardinal: direction
                        })));
                    }
                    if (i === 1)
                        break;
                    const newNeighbour = getNeighbour(neighbour.block, direction);
                    if (!newNeighbour || newNeighbour.type !== "conventional")
                        break;
                    neighbour = newNeighbour;
                }
                batchChanges.forEach(d => d()); // Ensures all changes happen only if everything succeeds
                return applyRailDetails(permutation, { cardinal: direction });
            }
        }
    // Generic rail handling
    const batchChanges = [];
    const checkDirections = [];
    for (const direction of allDirections) {
        const neighbour = getNeighbour(location, direction);
        if (!neighbour)
            continue;
        const railDirections = getRailDirections(neighbour.block, neighbour.block.permutation);
        if (!railDirections?.checked)
            continue;
        if ((ownDetails.type === "switch" && neighbour.type !== "conventional") || ownDetails.type !== "conventional" && neighbour.type === "switch")
            return false;
        checkDirections.push({ direction, neighbour, railDirections });
    }
    let markForDeath = false;
    const directionsOfDeathMarkers = [];
    let found = [];
    for (const { direction, neighbour, railDirections } of checkDirections) {
        // If a rail is connected to another rail via all its directions than we can't connect to it
        const usedConnectionDirections = railDirections.connectedTo.filter(d => railDirections[d] && railDirections[d].hasConnectedNeighbour);
        const availableConnectionDirections = railDirections.connectedTo.filter(d => railDirections[d] && !railDirections[d].hasConnectedNeighbour);
        if (!availableConnectionDirections.length) {
            directionsOfDeathMarkers.push(direction);
            markForDeath = true;
            continue;
        }
        const reciprocal = getReciprocal(direction);
        // Is the other rail already lined up for this connection
        const alreadyMatches = railDirections.connectedTo.some(d => d === reciprocal);
        // Get possible changes to force rotation if it is not
        let connectOther;
        if (!alreadyMatches) {
            // Slope rail cannot be rotated to fit new rail placement
            if (neighbour.type === "conventional" && neighbour.slope !== 0 && neighbour.slope < 4) {
                return false;
            }
            for (const connectionDirection of availableConnectionDirections) {
                const dir = railDirections[connectionDirection];
                if (!dir || dir.hasConnectedNeighbour)
                    continue;
                connectOther = rotate(usedConnectionDirections, reciprocal, neighbour);
                if (connectOther)
                    break;
            }
            if (!connectOther) {
                directionsOfDeathMarkers.push(direction);
                markForDeath = true;
                continue;
            }
        }
        // Determine how the rail being placed must rotate
        const connectSelf = rotate(found, direction, ownDetails);
        if (!connectSelf) {
            markForDeath = true;
            continue;
        }
        if (connectOther) {
            const currentNeighbour = neighbour;
            batchChanges.push(() => currentNeighbour.block.setPermutation(applyRailDetails(currentNeighbour.block.permutation, connectOther)));
        }
        permutation = applyRailDetails(permutation, connectSelf);
        // Slope join handling, for placing rails as part of an existing slope
        if (neighbour?.type === "conventional" && connectSelf.type === "conventional" &&
            neighbour.slope && !connectSelf.bend &&
            (direction === neighbour.cardinal || direction === getReciprocal(neighbour.cardinal))) {
            const newSlope = direction !== neighbour.cardinal ? neighbour.slope - 1 : neighbour.slope + 1;
            if (newSlope > 0 && newSlope < 4) {
                permutation = applyRailDetails(permutation, {
                    type: "conventional",
                    slope: newSlope,
                    cardinal: neighbour.cardinal
                });
                break;
            }
        }
        // This check ends the search if the rail has found its maximum number of possible connections
        if (ownDetails.type === "buffer")
            break;
        else if (ownDetails.type === "switch" && found.length === 2)
            break;
        else if (ownDetails.type === "conventional" && found.length === 1)
            break;
        found.push(direction);
    }
    // Marked for death causes failure if no valid connections are found
    if (!found.length && markForDeath)
        return false;
    // Determine invalid placement
    const { connectedTo } = getRailDirections(location, permutation, false);
    for (const connectionDirection of connectedTo) {
        for (const deathMarkerDirection of directionsOfDeathMarkers) {
            if (connectionDirection === deathMarkerDirection)
                return false;
        }
    }
    batchChanges.forEach(d => d());
    return permutation;
}
const rail = {
    beforeOnPlayerPlace(e) {
        const connectedRailPermutation = autoConnectRail(e.block, e.permutationToPlace);
        if (!connectedRailPermutation) {
            e.cancel = true;
            return;
        }
        e.permutationToPlace = connectedRailPermutation;
    }
};

const crossingSignal = {
    beforeOnPlayerPlace(e) {
        const above = e.block.above();
        if (!above?.isAir || !above.isValid()) {
            e.cancel = true;
            return;
        }
        above?.setType("pod_trn:crossing_signal_top");
    },
    onTick(e) {
        if (e.block.above()?.typeId !== "pod_trn:crossing_signal_top" ||
            e.block.north()?.hasTag("pod_trn:crossing_signal") ||
            e.block.east()?.hasTag("pod_trn:crossing_signal") ||
            e.block.south()?.hasTag("pod_trn:crossing_signal") ||
            e.block.west()?.hasTag("pod_trn:crossing_signal")) {
            e.dimension.getEntities({
                type: "minecraft:item",
                location: e.block,
                maxDistance: 10
            }).forEach(item => {
                const component = item.getComponent(EntityItemComponent.componentId);
                if (!component?.isValid())
                    return;
                if (component.itemStack.typeId !== "pod_trn:crossing_signal_top")
                    return;
                item.remove();
            });
            const itemStack = e.block.getItemStack();
            if (!itemStack)
                return;
            e.dimension.spawnItem(itemStack, e.block);
            e.block.setType("air");
            return;
        }
        e.block.setPermutation(e.block.permutation.withState("pod_trn:active", !!e.dimension.getEntities({
            families: ["pod_trn_locomotive"],
            location: e.block,
            maxDistance: 25
        }).length));
        if (!e.block.permutation.getState("pod_trn:active"))
            return;
        if (!e.block.permutation.getState("pod_trn:muted"))
            e.dimension.playSound("pod_trn.crossing_signal", e.block);
    }
};
const crossingSignalTop = {
    beforeOnPlayerPlace(e) {
        e.cancel = true;
        return;
    },
    onTick(e) {
        if (!e.block.below()?.hasTag("pod_trn:crossing_signal")) {
            e.block.setType("air");
            return;
        }
    }
};

function between(lower, value, higher) {
    return lower <= value && value <= higher;
}
const trestle = {
    beforeOnPlayerPlace(e) {
        for (const neighbour of [e.block.above(), e.block.below()]) {
            if (!neighbour?.isValid() || !neighbour?.hasTag("pod_trn:trestle"))
                continue;
            e.permutationToPlace = e.permutationToPlace
                .withState("minecraft:cardinal_direction", neighbour?.permutation.getState("minecraft:cardinal_direction"))
                .withState("pod_trn:trestle_direction", neighbour?.permutation.getState("pod_trn:trestle_direction"));
            return;
        }
        const { y } = e.player?.getRotation();
        const cardinal = e.permutationToPlace.getState("minecraft:cardinal_direction");
        let apply = "straight";
        if (cardinal === "north") {
            if (between(135, y, 150))
                apply = "left";
            else if (between(-150, y, -135))
                apply = "right";
        }
        else if (cardinal === "east") {
            if (between(-135, y, -110))
                apply = "left";
            else if (between(-70, y, -45))
                apply = "right";
        }
        else if (cardinal === "south") {
            if (between(-45, y, -20))
                apply = "left";
            else if (between(20, y, 45))
                apply = "right";
        }
        else if (cardinal === "west") {
            if (between(45, y, 70))
                apply = "left";
            else if (between(110, y, 135))
                apply = "right";
        }
        e.permutationToPlace = e.permutationToPlace.withState("pod_trn:trestle_direction", apply);
    }
};

function platformStates(block, permutation) {
    if (!block?.isValid())
        return null;
    permutation ??= block.permutation;
    if (!permutation.hasTag("pod_trn:platform"))
        return null;
    return {
        block,
        cardinal: getCardinal(permutation),
        shape: permutation.getState("pod_trn:shape")
    };
}
function update(block, permutation) {
    const { cardinal } = platformStates(block, permutation);
    const neighbours = {
        North: platformStates(block.north()),
        East: platformStates(block.east()),
        South: platformStates(block.south()),
        West: platformStates(block.west())
    };
    const leftDir = directionRelativeTo("West", cardinal);
    const rightDir = directionRelativeTo("East", cardinal);
    const right = neighbours[rightDir];
    const left = neighbours[leftDir];
    const back = neighbours[directionRelativeTo("North", cardinal)];
    const front = neighbours[directionRelativeTo("South", cardinal)];
    if (left?.cardinal === leftDir)
        left.block.setPermutation(left.block?.permutation.withState("pod_trn:shape", "inner_right"));
    else if (left?.cardinal === rightDir)
        left.block.setPermutation(left.block?.permutation.withState("pod_trn:shape", "outer_left"));
    if (right?.cardinal === leftDir)
        right.block.setPermutation(right.block?.permutation.withState("pod_trn:shape", "outer_right"));
    else if (right?.cardinal === rightDir)
        right.block.setPermutation(right.block?.permutation.withState("pod_trn:shape", "inner_left"));
    if (back) {
        if (back.cardinal === leftDir && right?.cardinal !== cardinal)
            return permutation.withState("pod_trn:shape", "outer_left");
        if (back.cardinal === rightDir && left?.cardinal !== cardinal)
            return permutation.withState("pod_trn:shape", "outer_right");
    }
    if (front) {
        if (front.cardinal === leftDir && left?.cardinal !== cardinal)
            return permutation.withState("pod_trn:shape", "inner_left");
        if (front.cardinal === rightDir && right?.cardinal !== cardinal)
            return permutation.withState("pod_trn:shape", "inner_right");
    }
    return permutation;
}
const platform = {
    beforeOnPlayerPlace(e) {
        e.permutationToPlace = update(e.block, e.permutationToPlace);
    }
};

const TrainData = {
    "pod_trn:steam_locomotive": {
        stoppingDistance: 3,
        horn: "pod_trn.steam.horn",
        roll: "pod_trn.chug",
        startStopSound: "random.fizz",
        initialVariant: 7,
        speedMultiplier: 1,
        canHitch: true,
        canShunt: true,
        capacity: 15,
        usesFuel: {
            usage: [1 / 24000, 1 / 12000],
            sources: [
                { typeId: "minecraft:coal_block", fills: 1 },
                { typeId: "minecraft:coal", fills: 1 / 9 },
                { typeId: "minecraft:charcoal", fills: 1 / 9 }
            ]
        }
    },
    "pod_trn:diesel_locomotive": {
        stoppingDistance: 3,
        initialVariant: 11,
        horn: "pod_trn.air.horn",
        roll: "pod_trn.diesel",
        speedMultiplier: 0.8,
        canHitch: true,
        canShunt: true,
        capacity: 30,
        usesFuel: {
            usage: [1 / 48000, 1 / 24000],
            sources: [
                { typeId: "pod_trn:fuel_bucket", fills: 1 / 2, returnsBucket: true },
            ]
        }
    },
    "pod_trn:passenger_car": {
        stoppingDistance: 3,
        initialVariant: 9,
        pullInFamilies: ["villager", "piglin"],
        canHitch: true
    },
    "pod_trn:caboose": {
        stoppingDistance: 3,
        initialVariant: 0,
        pullInFamilies: ["villager", "piglin"],
        canHitch: true
    },
    "pod_trn:animal_car": {
        stoppingDistance: 3,
        pullInFamilies: ["mob", "monster"],
        canHitch: true
    },
    "pod_trn:flatbed_car": {
        stoppingDistance: 3,
        canHitch: true
    },
    "pod_trn:gondola_car": {
        stoppingDistance: 3,
        canHitch: true
    },
    "pod_trn:handcar": {
        stoppingDistance: 1,
        initialVariant: 13
    },
    default: {
        initialVariant: 10,
        stoppingDistance: 1,
        speedMultiplier: 0.5,
        canHitch: false,
        canShunt: false
    }
};
const StackContainerCapacity = 5;
const GondolaItems = {
    "minecraft:wheat": {
        cargoType: "grain",
        colour: { r: 0.65, g: 0.53, b: 0.25 }
    },
    "minecraft:wheat_seeds": {
        cargoType: "grain",
        colour: { r: 0.35, g: 0.53, b: 0.06 }
    },
    "minecraft:beetroot_seeds": {
        cargoType: "grain",
        colour: { r: 0.6, g: 0.42, b: 0.19 }
    },
    "minecraft:melon_seeds": {
        cargoType: "grain",
        colour: { r: 0.15, g: 0.11, b: 0.03 }
    },
    "minecraft:pumpkin_seeds": {
        cargoType: "grain",
        colour: { r: 0.8, g: 0.7, b: 0.43 }
    },
    "minecraft:torchflower_seeds": {
        cargoType: "grain",
        colour: { r: 0.52, g: 0.54, b: 0.3 }
    },
    "minecraft:cocoa_beans": {
        cargoType: "grain",
        colour: { r: 0.29, g: 0.17, b: 0.07 }
    },
    "minecraft:nether_wart": {
        cargoType: "grain",
        colour: { r: 0.43, g: 0.08, b: 0.11 }
    },
    "minecraft:bone_meal": {
        cargoType: "gravel",
        colour: { r: 0.7, g: 0.7, b: 0.75 }
    },
    "minecraft:hay_block": {
        convert: true,
        to: "minecraft:wheat",
        amount: 9
    },
    "minecraft:bone_block": {
        convert: true,
        to: "minecraft:bone_meal",
        amount: 9
    },
    "minecraft:charcoal": {
        cargoType: "stone",
        colour: { r: 0.04, g: 0.03, b: 0.01 }
    },
    "minecraft:coal": {
        cargoType: "stone",
        colour: { r: 0, g: 0, b: 0 }
    },
    "minecraft:raw_iron": {
        cargoType: "mineral",
        colour: { r: 0.49, g: 0.33, b: 0.22 }
    },
    "minecraft:raw_copper": {
        cargoType: "mineral",
        colour: { r: 0.65, g: 0.29, b: 0.15 }
    },
    "minecraft:raw_gold": {
        cargoType: "mineral",
        colour: { r: 0.89, g: 0.62, b: 0.05 }
    },
    "minecraft:redstone": {
        cargoType: "gravel",
        colour: { r: 0.63, g: 0.02, b: 0 }
    },
    "minecraft:lapis_lazuli": {
        cargoType: "mineral",
        colour: { r: 0.04, g: 0.23, b: 0.56 }
    },
    "minecraft:diamond": {
        cargoType: "mineral",
        colour: { r: 0.12, g: 0.77, b: 0.71 }
    },
    "minecraft:emerald": {
        cargoType: "mineral",
        colour: { r: 0, g: 0.67, b: 0.17 }
    },
    "minecraft:amethyst_shard": {
        cargoType: "mineral",
        colour: { r: 0.23, g: 0.11, b: 0.47 }
    },
    "minecraft:quartz": {
        cargoType: "mineral",
        colour: { r: 0.87, g: 0.79, b: 0.73 }
    },
    "minecraft:coal_block": {
        convert: true,
        to: "minecraft:coal",
        amount: 9
    },
    "minecraft:redstone_block": {
        convert: true,
        to: "minecraft:redstone",
        amount: 9
    },
    "minecraft:lapis_block": {
        convert: true,
        to: "minecraft:lapis_lazuli",
        amount: 9
    },
    "minecraft:raw_iron_block": {
        convert: true,
        to: "minecraft:raw_iron",
        amount: 9
    },
    "minecraft:raw_copper_block": {
        convert: true,
        to: "minecraft:raw_iron",
        amount: 9
    },
    "minecraft:raw_gold_block": {
        convert: true,
        to: "minecraft:raw_gold",
        amount: 9
    },
    "minecraft:diamond_block": {
        convert: true,
        to: "minecraft:diamond",
        amount: 9
    },
    "minecraft:emerald_block": {
        convert: true,
        to: "minecraft:emerald",
        amount: 9
    },
    "minecraft:clay_ball": {
        cargoType: "gravel",
        colour: { r: 0.46, g: 0.49, b: 0.56 }
    },
    "minecraft:gravel": {
        cargoType: "gravel",
        colour: { r: 0.37, g: 0.35, b: 0.35 }
    },
    "minecraft:sand": {
        cargoType: "gravel",
        colour: { r: 0.82, g: 0.73, b: 0.54 }
    },
    "minecraft:dirt": {
        cargoType: "gravel",
        colour: { r: 0.34, g: 0.19, b: 0.08 }
    },
    "minecraft:red_sand": {
        cargoType: "gravel",
        colour: { r: 0.67, g: 0.31, b: 0.02 }
    },
    "minecraft:netherrack": {
        cargoType: "stone",
        colour: { r: 0.28, g: 0.04, b: 0.04 }
    },
    "minecraft:cobblestone": {
        cargoType: "stone",
        colour: { r: 0.28, g: 0.28, b: 0.28 }
    },
    "minecraft:cobbled_deepslate": {
        cargoType: "stone",
        colour: { r: 0.11, g: 0.11, b: 0.12 }
    },
    "minecraft:blackstone": {
        cargoType: "stone",
        colour: { r: 0.03, g: 0.02, b: 0.05 }
    },
    "minecraft:white_concrete_powder": {
        cargoType: "gravel",
        colour: { r: 0.7, g: 0.7, b: 0.7 }
    },
    "minecraft:light_gray_concrete_powder": {
        cargoType: "gravel",
        colour: { r: 0.44, g: 0.43, b: 0.42 }
    },
    "minecraft:gray_concrete_powder": {
        cargoType: "gravel",
        colour: { r: 0.17, g: 0.18, b: 0.19 }
    },
    "minecraft:black_concrete_powder": {
        cargoType: "gravel",
        colour: { r: 0, g: 0, b: 0 }
    },
    "minecraft:brown_concrete_powder": {
        cargoType: "gravel",
        colour: { r: 0.28, g: 0.16, b: 0.07 }
    },
    "minecraft:red_concrete_powder": {
        cargoType: "gravel",
        colour: { r: 0.49, g: 0.06, b: 0.05 }
    },
    "minecraft:orange_concrete_powder": {
        cargoType: "gravel",
        colour: { r: 0.72, g: 0.35, b: 0 }
    },
    "minecraft:yellow_concrete_powder": {
        cargoType: "gravel",
        colour: { r: 0.78, g: 0.61, b: 0.13 }
    },
    "minecraft:lime_concrete_powder": {
        cargoType: "gravel",
        colour: { r: 0.31, g: 0.55, b: 0 }
    },
    "minecraft:green_concrete_powder": {
        cargoType: "gravel",
        colour: { r: 0.20, g: 0.26, b: 0.04 }
    },
    "minecraft:cyan_concrete_powder": {
        cargoType: "gravel",
        colour: { r: 0, g: 0.43, b: 0.44 }
    },
    "minecraft:light_blue_concrete_powder": {
        cargoType: "gravel",
        colour: { r: 0.13, g: 0.54, b: 0.68 }
    },
    "minecraft:blue_concrete_powder": {
        cargoType: "gravel",
        colour: { r: 0.11, g: 0.12, b: 0.44 }
    },
    "minecraft:purple_concrete_powder": {
        cargoType: "gravel",
        colour: { r: 0.32, g: 0.09, b: 0.44 }
    },
    "minecraft:magenta_concrete_powder": {
        cargoType: "gravel",
        colour: { r: 0.5, g: 0.17, b: 0.47 }
    },
    "minecraft:pink_concrete_powder": {
        cargoType: "gravel",
        colour: { r: 0.72, g: 0.4, b: 0.51 }
    }
};
// The order of this variable matters, be careful when changing the order as this will corrupt existing flatbeds
const FlatbedCargoTypes = [
    "jungle",
    "mangrove",
    "oak",
    "dark_oak",
    "spruce",
    "acacia",
    "birch",
    "cherry",
    "crimson",
    "warped",
    "stripped_jungle",
    "stripped_mangrove",
    "stripped_oak",
    "stripped_dark_oak",
    "stripped_spruce",
    "stripped_acacia",
    "stripped_birch",
    "stripped_cherry",
    "stripped_crimson",
    "stripped_warped",
    "bamboo",
    "stripped_bamboo",
    "brick",
    "pale_oak",
    "stripped_pale_oak",
];
const FlatbedItems = {
    "minecraft:jungle_log": {
        cargoType: "jungle"
    },
    "minecraft:mangrove_log": {
        cargoType: "mangrove"
    },
    "minecraft:oak_log": {
        cargoType: "oak"
    },
    "minecraft:pale_oak_log": {
        cargoType: "pale_oak"
    },
    "minecraft:spruce_log": {
        cargoType: "spruce"
    },
    "minecraft:dark_oak_log": {
        cargoType: "dark_oak"
    },
    "minecraft:acacia_log": {
        cargoType: "acacia"
    },
    "minecraft:birch_log": {
        cargoType: "birch"
    },
    "minecraft:cherry_log": {
        cargoType: "cherry"
    },
    "minecraft:crimson_stem": {
        cargoType: "crimson"
    },
    "minecraft:warped_stem": {
        cargoType: "warped"
    },
    "minecraft:bamboo_block": {
        cargoType: "bamboo"
    },
    "minecraft:stripped_jungle_log": {
        cargoType: "stripped_jungle"
    },
    "minecraft:stripped_mangrove_log": {
        cargoType: "stripped_mangrove"
    },
    "minecraft:stripped_oak_log": {
        cargoType: "stripped_oak"
    },
    "minecraft:stripped_pale_oak_log": {
        cargoType: "stripped_pale_oak"
    },
    "minecraft:stripped_spruce_log": {
        cargoType: "stripped_spruce"
    },
    "minecraft:stripped_dark_oak_log": {
        cargoType: "stripped_dark_oak"
    },
    "minecraft:stripped_acacia_log": {
        cargoType: "stripped_acacia"
    },
    "minecraft:stripped_birch_log": {
        cargoType: "stripped_birch"
    },
    "minecraft:stripped_cherry_log": {
        cargoType: "stripped_cherry"
    },
    "minecraft:stripped_crimson_stem": {
        cargoType: "stripped_crimson"
    },
    "minecraft:stripped_warped_stem": {
        cargoType: "stripped_warped"
    },
    "minecraft:stripped_bamboo_block": {
        cargoType: "stripped_bamboo"
    },
    "minecraft:brick_block": {
        cargoType: "brick"
    }
};
function getHorn(id) {
    return TrainData[id]?.horn ?? TrainData.default.horn;
}
function getStartStopSound(id) {
    return TrainData[id]?.startStopSound ?? TrainData.default.startStopSound;
}
function getRoll(id) {
    return TrainData[id]?.roll ?? TrainData.default.roll ?? "";
}
function getCanHitch(id) {
    return TrainData[id]?.canHitch ?? TrainData.default.canHitch;
}
function getCanShunt(id) {
    return TrainData[id]?.canShunt ?? TrainData.default.canShunt;
}
function getSpeedMultiplier(id) {
    return TrainData[id]?.speedMultiplier ?? TrainData.default.speedMultiplier;
}
function getStoppingDistance(id) {
    return TrainData[id]?.stoppingDistance ?? TrainData.default.stoppingDistance;
}
function getCapacity(id) {
    return TrainData[id]?.capacity ?? 0;
}
function getInitialVariant(id) {
    return TrainData[id]?.initialVariant ?? TrainData.default.initialVariant ?? 0;
}
function getPullInFamilies(id) {
    return TrainData[id]?.pullInFamilies ?? TrainData.default.pullInFamilies ?? [];
}
function getUsesFuel(id) {
    return TrainData[id]?.usesFuel;
}
const variants = [
    "red",
    "blue",
    "cyan",
    "gray",
    "light_gray",
    "lime",
    "pink",
    "black",
    "brown",
    "green",
    "white",
    "orange",
    "purple",
    "yellow",
    "magenta",
    "light_blue",
];

const directionRotations = [
    ["North", 0],
    ["NorthEast", 45],
    ["East", 90],
    ["SouthEast", 135],
    ["South", 180],
    ["SouthWest", 225],
    ["West", 270],
    ["NorthWest", 305],
];
function railDirectionToVector2(direction) {
    switch (direction) {
        case "North":
            return { x: 0, y: 0 };
        case "NorthEast":
            return { x: 0, y: 45 };
        case "East":
            return { x: 0, y: 90 };
        case "SouthEast":
            return { x: 0, y: 135 };
        case "South":
            return { x: 0, y: 180 };
        case "SouthWest":
            return { x: 0, y: -135 };
        case "West":
            return { x: 0, y: -90 };
        case "NorthWest":
            return { x: 0, y: -45 };
    }
}
function mod(value, mod) {
    return ((value % mod) + mod) % mod;
}
function yRotationToDirection(yRot) {
    const normalRotation = mod(yRot, 360);
    let min = Infinity;
    let foundDir = undefined;
    for (const [dir, rot] of directionRotations) {
        const difference = Math.abs(normalRotation - rot);
        if (difference < min) {
            min = difference;
            foundDir = dir;
        }
    }
    return foundDir;
}
function interpolateDegrees(value, target, step) {
    const diffA = target - value;
    const diffB = target - 360 - value;
    if (Math.abs(diffA) < Math.abs(diffB)) {
        if (Math.abs(diffA) < 1.5)
            return target;
        return value + diffA * step;
    }
    if (Math.abs(diffB) < 1.5)
        return target;
    return value + diffB * step;
}

const Properties = {
    Name: "name",
    Throttle: "speed",
    ThrottleWas: "speed_was",
    Horn: "auto_horn",
    RespectSignals: "respect_signals",
    LookThrottle: "look_throttle",
};
const THROTTLE_MIN = -5;
const THROTTLE_MAX = 10;
const LOOK_THRESHOLD = 0.15;
const DELTA = 1 / 20;
const SLOPE_ANGLE = 14.036;
const VERT_SPEED_MULTIPLIER = 1;
const DESIRED_FLAT_Y_LEVEL = 0.2;
const DESIRED_SLOPE_1_Y_LEVEL_MAX = 0.44;
const LOOKAHEAD_AMOUNT = 5;
function splitCoordinates(location) {
    const wholePart = { ...location };
    const decimalPart = { ...location };
    wholePart.x = Math.floor(wholePart.x);
    decimalPart.x -= wholePart.x;
    wholePart.y = Math.floor(wholePart.y);
    decimalPart.y -= wholePart.y;
    wholePart.z = Math.floor(wholePart.z);
    decimalPart.z -= wholePart.z;
    return [wholePart, decimalPart];
}
function calculateCapacityThrottle(e, throttle, carsAttached) {
    const capacity = getCapacity(e.typeId);
    const use = 1 - carsAttached / capacity;
    return capacity === 0 ? throttle : throttle * use;
}
function locomotiveProcess(e) {
    if (!e?.isValid())
        return;
    const settings = getSettings(e);
    const rider = getControllingRider(e);
    const horn = getHorn(e.typeId);
    if (rider?.isValid()) {
        // Horn or jump throttle
        if (horn) {
            if (rider.isJumping)
                e.dimension.playSound(horn, e.location, { volume: 3 });
        }
        else {
            setSettings(e, { speed: (rider?.isJumping ? 5 : 0) * (!e.getDynamicProperty("invert_handcar") ? 1 : -1) });
        }
        // Look controls
        if (settings[Properties.LookThrottle]) {
            const look = rider?.getViewDirection()?.y;
            const lookPercent = Math.min((Math.abs(look) - LOOK_THRESHOLD) / 0.4, 1);
            if (look > LOOK_THRESHOLD) {
                setSettings(e, { speed: Math.floor(THROTTLE_MAX * lookPercent) });
            }
            else if (look < -LOOK_THRESHOLD) {
                setSettings(e, { speed: Math.floor(THROTTLE_MIN * lookPercent) });
            }
            else {
                setSettings(e, { speed: 0 });
            }
        }
    }
    else {
        // Hack to stop handcar
        if (!horn) {
            setSettings(e, { speed: 0 });
        }
    }
    const canShunt = getCanShunt(e.typeId);
    const throttle = settings[Properties.Throttle];
    const usesFuel = getUsesFuel(e.typeId);
    if (usesFuel && throttle) {
        const [min, max] = usesFuel.usage;
        const speedPercent = Math.abs(throttle) / 10;
        const usage = (max - min) * speedPercent + min;
        e.setProperty("pod_trn:fuel", Math.max(0, e.getProperty("pod_trn:fuel") - usage));
    }
    const firstBatch = trainSetProcess(e, throttle);
    if (!firstBatch)
        return;
    if (!canShunt || throttle < 0) {
        runBatch(firstBatch, calculateCapacityThrottle(e, throttle, firstBatch.length), getSpeedMultiplier(e.typeId));
        return;
    }
    // Push cars ahead
    const secondBatch = trainSetProcess(e, throttle, true);
    if (!secondBatch)
        return;
    const [_to, by] = getHitchInfo(e);
    const speed = calculateCapacityThrottle(e, throttle, secondBatch.length) * getSpeedMultiplier(e.typeId);
    if (by?.isValid() && speed < by.getDynamicProperty("last_capacity_throttle")) {
        // log("unhitched1")
        unhitch(e, by);
    }
    runBatch([...firstBatch, ...secondBatch.slice(1)], calculateCapacityThrottle(e, throttle, firstBatch.length + secondBatch.length - 1), getSpeedMultiplier(e.typeId));
}
function carProcess(e) {
    const batch = trainSetProcess(e, 0);
    if (!batch)
        return;
    runBatch(batch);
}
function runBatch(entities, throttle = 0, multiplier = 0) {
    let batch = [];
    let haltFlag = false;
    for (const entity of entities) {
        const increment = trainProcess(entity, throttle, multiplier);
        if (!increment) {
            haltFlag = true;
            batch.push({ entity, impulse: false });
            continue;
        }
        batch.push(increment);
    }
    for (const increment of batch) {
        const { impulse, entity } = increment;
        if (!entity.isValid())
            return;
        entity.clearVelocity();
        if (!impulse || !throttle || haltFlag)
            continue;
        if (!impulse || !throttle || haltFlag) {
            entity.setDynamicProperty("last_capacity_throttle", 0);
            continue;
        }
        const { rotation, entitiesHit } = increment;
        entity.setDynamicProperty("last_capacity_throttle", throttle * multiplier);
        entity.setRotation(rotation);
        entity.applyImpulse(impulse);
        entity.setProperty("pod_trn:direction", throttle > 0 ? 1 : -1);
        const divisor = Math.floor((TicksPerSecond * 4 / (Math.abs(throttle) * multiplier)));
        if (system.currentTick % divisor === 0) {
            entity.dimension.playSound(getRoll(entity.typeId), entity.location, {
                pitch: 1,
                volume: 2
            });
        }
        if ((system.currentTick + Math.floor(Math.abs(divisor / 2))) % divisor === 0) {
            entity.dimension.playSound(getRoll(entity.typeId), entity.location, {
                pitch: 0.97,
                volume: 2
            });
        }
        if (system.currentTick % Math.max(Math.floor(divisor * 0.15), 5) === 0) {
            entity.dimension.playSound("pod_trn.roll", entity.location, {
                pitch: 1,
                volume: 2
            });
        }
        const launchSpeed = throttle * 3 * DELTA;
        for (const hitEntity of entitiesHit) {
            if (!hitEntity.isValid())
                continue;
            hitEntity.applyKnockback(impulse.x + Math.random() * 0.2 - 0.1, impulse.z + Math.random() * 0.2 - 0.1, launchSpeed, launchSpeed);
        }
    }
}
function trainSetProcess(initialEntity, throttle = 0, shunting = false) {
    const entities = [];
    const canShunt = getCanShunt(initialEntity.typeId);
    if (!canShunt && getEntity(initialEntity.getDynamicProperty("hitched_by")))
        return false;
    const initialIsLocomotive = initialEntity.matches({ families: ["pod_trn_locomotive"] });
    const forwardsOrNil = throttle >= 0;
    let currentEntity = initialEntity;
    let count = 0; // Prevent infinite loops
    while (count++ < 32) {
        if (!currentEntity.isValid())
            return false;
        entities.push(currentEntity);
        const hitches = getHitchInfo(currentEntity);
        const hitchedEntity = hitches[shunting ? 1 : 0];
        if (hitchedEntity === null)
            break;
        if (hitchedEntity === undefined)
            return false;
        if (hitchedEntity.matches({ families: ["pod_trn_locomotive"] }) && forwardsOrNil) {
            const hitchedSpeed = calculateCapacityThrottle(hitchedEntity, getSettings(hitchedEntity)[Properties.Throttle], count) * getSpeedMultiplier(hitchedEntity.typeId);
            const initialEntitySpeed = calculateCapacityThrottle(initialEntity, getSettings(initialEntity)[Properties.Throttle], count) * getSpeedMultiplier(initialEntity.typeId);
            // Locomotive is being pulled or pushed
            if (!shunting && initialIsLocomotive && hitchedSpeed < initialEntitySpeed) {
                unhitch(hitchedEntity, currentEntity);
                // log("Decouple because faster")
                break;
            }
            // Non-locomotives and locomotives with no throttle should concede control to shunting locomotive
            if (!shunting && throttle === 0)
                return false;
            // Shunting is only allowed if none of the locomotives ahead have throttle engaged
            if (shunting && hitchedSpeed !== 0)
                return false;
        }
        currentEntity = hitchedEntity;
    }
    return entities;
}
const STATION_DELAYS = [0, TicksPerSecond * 5, TicksPerSecond * 10, TicksPerSecond * 20, TicksPerSecond * 60];
function trainProcess(e, throttle, multiplier) {
    try {
        const rail = getEntityRail(e);
        if (!rail?.checked) {
            const derailment = e.getDynamicProperty("derailment") ?? 0;
            if (derailment > 1) {
                popTrain(e);
                return;
            }
            e.setDynamicProperty("derailment", derailment + 1);
            return;
        }
        e.setDynamicProperty("derailment", undefined);
        // Pull in entities
        const pullIn = getPullInFamilies(e.typeId);
        if (pullIn.length) {
            const riders = e.getComponent(EntityRideableComponent.componentId);
            if (riders.isValid()) {
                for (const family of pullIn) {
                    for (const potential of e.dimension.getEntities({ location: e.location, maxDistance: 3, families: [family], closest: 2, excludeFamilies: ["inanimate"] })) {
                        const riding = potential.getComponent(EntityRidingComponent.componentId);
                        if (riding?.isValid())
                            continue;
                        riders.addRider(potential);
                    }
                }
            }
        }
        // Pick up items (gondola)
        if (e.typeId === "pod_trn:gondola_car") {
            const itemEntities = e.dimension.getEntities({ location: e.location, maxDistance: 1, type: "item" });
            for (const itemEntity of itemEntities) {
                const itemComponent = itemEntity.getComponent(EntityItemComponent.componentId);
                if (!itemComponent?.isValid())
                    continue;
                const added = addItemToStackContainerCar(e, itemComponent.itemStack);
                if (!added)
                    continue;
                if (added < itemComponent.itemStack.amount) {
                    const copy = itemComponent.itemStack.clone();
                    copy.amount -= added;
                    e.dimension.spawnItem(copy, e.location);
                }
                itemEntity.remove();
            }
        }
        // Markers
        const blockBelow = rail.block.below();
        if (blockBelow?.isValid() && e.matches({ families: ["pod_trn_locomotive"], excludeTypes: ["pod_trn:handcar"] })) {
            if (blockBelow.typeId === "pod_trn:station_marker") {
                const value = blockBelow.permutation.getState("pod_trn:delay");
                const tickDelay = e.getDynamicProperty(`speed_swap_back_delay`);
                const throttleWas = e.getDynamicProperty(Properties.ThrottleWas);
                if (typeof tickDelay === "number" && tickDelay <= system.currentTick && !blockBelow?.permutation.getState("pod_trn:powered")) {
                    e.setDynamicProperty(`speed_swap_back_delay`, undefined);
                    e.setDynamicProperty(Properties.ThrottleWas, undefined);
                    setSettings(e, { speed: throttleWas });
                    blockBelow?.setPermutation(blockBelow?.permutation.withState("pod_trn:detect", false));
                }
                else if (!vectorsEqual(blockBelow, e.getDynamicProperty("remember_trigger"))) {
                    if (blockBelow?.permutation.getState("pod_trn:powered")) {
                        setSettings(e, { speed: 0 }, TicksPerSecond / 2);
                        throttle = 0;
                    }
                    else if (value) {
                        setSettings(e, { speed: 0 }, STATION_DELAYS[value]);
                        throttle = 0;
                    }
                    else {
                        system.runTimeout(() => {
                            try {
                                blockBelow?.setPermutation(blockBelow?.permutation.withState("pod_trn:detect", false));
                            }
                            catch (e) {
                                // Silently ignore error
                            }
                        });
                    }
                    e.setDynamicProperty("remember_trigger", blockBelow);
                    blockBelow?.setPermutation(blockBelow?.permutation.withState("pod_trn:detect", true));
                }
            }
            else {
                e.setDynamicProperty("remember_trigger", undefined);
            }
        }
        const wrongDifference = throttle !== 0 ? validateHitch(e) / 4 : 0; // Divide to not make the jump too extreme
        return calculateEntityIncrement(e, rail, throttle * multiplier * DELTA + wrongDifference);
    }
    catch (e) {
        // Invalid rail block
    }
}
function getSettings(e) {
    const canMove = !getUsesFuel(e.typeId) ? true : e.getProperty("pod_trn:fuel") > 0;
    return {
        [Properties.Name]: e.getDynamicProperty(Properties.Name) ?? "Player's train",
        [Properties.Throttle]: canMove ? e.getDynamicProperty(Properties.Throttle) ?? 0 : 0,
        [Properties.ThrottleWas]: e.getDynamicProperty(Properties.ThrottleWas) ?? 0,
        [Properties.Horn]: e.getDynamicProperty(Properties.Horn) ?? true,
        [Properties.RespectSignals]: e.getDynamicProperty(Properties.RespectSignals) ?? true,
        [Properties.LookThrottle]: e.getDynamicProperty(Properties.LookThrottle) ?? false,
    };
}
function setSettings(e, settings, speedSwapBackDelay) {
    if (settings.name !== undefined)
        e.setDynamicProperty(Properties.Name, settings.name);
    if (settings.speed !== undefined) {
        const current = e.getDynamicProperty(Properties.Throttle);
        if (speedSwapBackDelay) {
            e.setDynamicProperty(Properties.ThrottleWas, current);
            if (speedSwapBackDelay >= 0)
                e.setDynamicProperty(`speed_swap_back_delay`, speedSwapBackDelay + system.currentTick);
        }
        else {
            e.setDynamicProperty(Properties.ThrottleWas, undefined);
        }
        if (e.getProperty("pod_trn:fuel") && ((current !== 0 && settings.speed === 0) || (!current && settings.speed !== 0))) {
            const startStopSound = getStartStopSound(e.typeId);
            if (startStopSound)
                e.dimension.playSound(startStopSound, e.location, { pitch: 0.5 });
        }
        e.setDynamicProperty(Properties.Throttle, settings.speed);
        e.setProperty("pod_trn:throttle", Math.abs(settings.speed) / THROTTLE_MAX);
    }
    if (settings.autoHorn !== undefined)
        e.setDynamicProperty(Properties.Horn, settings.autoHorn);
    if (settings.respectSignals !== undefined)
        e.setDynamicProperty(Properties.RespectSignals, settings.respectSignals);
    if (settings.lookThrottle !== undefined)
        e.setDynamicProperty(Properties.LookThrottle, settings.lookThrottle);
}
function getEntityRail(e) {
    function checkBlock(block, requires) {
        if (!block?.isValid())
            throw new Error("Invalid");
        if (block?.hasTag("pod_trn:rail")) {
            const directions = getRailDirections(block, block.permutation);
            directions.block = block;
            if (requires) {
                if (!directions?.details)
                    return;
                for (const [key, value] of Object.entries(requires)) {
                    if (directions.details[key] !== value)
                        return null;
                }
            }
            return directions;
        }
    }
    const blockOn = e.dimension.getBlock(e.location);
    let directions = checkBlock(blockOn);
    if (directions !== undefined)
        return directions;
    // Check below for level 3 slope
    directions = checkBlock(blockOn?.below(), { type: "conventional", slope: 3 });
    if (directions !== undefined)
        return directions;
    // Check above for level 4 slope
    directions = checkBlock(blockOn?.above(), { type: "conventional", slope: 4 });
    if (directions !== undefined)
        return directions;
    return null;
}
function popTrain(e, noItem = false) {
    try {
        if (!e.matches({ families: ["pod_trn_stock"] }))
            return;
        if (!noItem) {
            const item = new ItemStack(e.typeId);
            const variant = e.getComponent(EntityVariantComponent.componentId);
            item.setDynamicProperty("variant", variant.value);
            // Drop gondola items
            if (e.typeId === "pod_trn:gondola_car" || e.typeId === "pod_trn:flatbed_car") {
                const count = e.getDynamicProperty("count") ?? 0;
                const currentTypeId = count ? e.getDynamicProperty("type_id") : undefined;
                if (currentTypeId) {
                    const item = new ItemStack(currentTypeId);
                    let remaining = count;
                    while (remaining) {
                        item.amount = Math.min(item.maxAmount, remaining);
                        remaining -= item.amount;
                        e.dimension.spawnItem(item, e.location);
                    }
                }
            }
            e.dimension.spawnItem(item, e.location);
        }
        // Clear hitches
        const [to, by] = getHitchInfo(e);
        to?.setDynamicProperty("hitched_by", undefined);
        by?.setDynamicProperty("hitched_to", undefined);
        // Play break sound
        e.dimension.playSound("block.lantern.break", e.location, { pitch: 0.2 });
        const rememberTrigger = e.getDynamicProperty("remember_trigger");
        if (!rememberTrigger)
            return e.remove();
        const block = e.dimension.getBlock(rememberTrigger);
        if (block?.typeId !== "pod_trn:station_marker")
            return e.remove();
        block.setPermutation(block.permutation.withState("pod_trn:detect", false));
        e.remove();
    }
    catch (e) {
        // Likely caused by unloaded access, can be ignored
    }
}
function calculateVelocity(power, { x, y }) {
    const yRads = y * Math.PI / 180;
    const xRads = x * Math.PI / 180;
    const output = { x: 0, y: 0, z: 0 };
    // Rotate around x
    output.z = power * Math.cos(xRads);
    output.y = -power * Math.sin(xRads);
    // Rotate around y
    output.x = -output.z * Math.sin(yRads);
    output.z = output.z * Math.cos(yRads);
    // Lower vertical speed to correct levels
    output.y *= VERT_SPEED_MULTIPLIER;
    return output;
}
function traverseRail(block, hailingFrom, limit) {
    const findings = [];
    for (let i = 0; i < (limit ?? LOOKAHEAD_AMOUNT); i++) {
        const newDirections = getRailDirections(block, block?.permutation, true, hailingFrom);
        if (!newDirections?.checked)
            break;
        const heading = newDirections.connectedTo.filter(d => newDirections[d] && newDirections[d].traversable).at(0);
        if (!heading)
            break;
        const connection = newDirections[heading];
        if (!connection.block?.isValid())
            break;
        block = connection.block;
        hailingFrom = heading;
        const below = block.below();
        const above = block.above();
        const entities = block.permutation.getState("pod_trn:slope") === 4 && below?.isValid() ? block.dimension.getEntitiesAtBlockLocation(below) : block.dimension.getEntitiesAtBlockLocation(block);
        findings.push([hailingFrom, connection, entities.filter(e => e.isValid()), !above?.isAir || !above?.above()?.isAir]);
    }
    return findings;
}
function getEntitiesAhead(id, stoppingDistance, lookAhead, families, ignoreId) {
    const findings = lookAhead.map((d, index) => {
        return d[2].filter(d => {
            if (d.id === id || !d.matches({ families }))
                return false;
            const coupledModifier = d.id === ignoreId ? 2.5 : 0;
            const localStoppingDistance = getStoppingDistance(d.typeId);
            return stoppingDistance - coupledModifier >= index && localStoppingDistance - coupledModifier >= index;
        });
    }).flat();
    if (!findings.length)
        return false;
    return findings;
}
function isHeadingAllowed(e, forwards = true, heading, details) {
    const lookAhead = traverseRail(details?.block, heading);
    lookAhead.unshift([heading, details, []]);
    const stoppingDistance = getStoppingDistance(e.typeId);
    // Detect if there's a train entity that will collide on current heading
    const trainAhead = getEntitiesAhead(e.id, stoppingDistance, lookAhead, ["pod_trn_stock"], e.getDynamicProperty(forwards ? "hitched_by" : "hitched_to"));
    if (trainAhead) {
        if (trainAhead.length !== 1 || !trainAhead[0].matches({ families: ["pod_trn_stock"] }))
            return false;
        if (!forwards) {
            // if ((e.getDynamicProperty("last_known_amount") as number) >= (trainAhead[0].getDynamicProperty("last_known_amount") as number))
            //     return false
            if (!hitch(trainAhead[0], e))
                return false;
        }
        if (forwards) {
            // if ((e.getDynamicProperty("last_known_amount") as number) <= (trainAhead[0].getDynamicProperty("last_known_amount") as number))
            //     return false
            if (!hitch(e, trainAhead[0]))
                return false;
        }
    }
    // Check if rail ends, there's a buffer, or other blockage
    const bufferFound = lookAhead.slice(0, stoppingDistance).some(d => d[3] || d[1].block?.hasTag("pod_trn:buffer_rail"));
    if (bufferFound || lookAhead.length < stoppingDistance) {
        return false;
    }
    return lookAhead;
}
function isPlacementSafe(block, railDirections, typeId) {
    if (!railDirections?.checked ||
        railDirections.details.type !== "conventional" ||
        railDirections.details.bend ||
        railDirections.details.slope > 0)
        return false;
    const stoppingDistance = getStoppingDistance(typeId);
    if (block.dimension.getEntitiesAtBlockLocation(block).some(d => d.matches({ families: ["pod_trn_stock"] })))
        return false;
    return !railDirections.connectedTo
        .some((d) => {
        const lookAhead = traverseRail(block, d, stoppingDistance);
        if (lookAhead.some(d => !d[1].block?.hasTag("pod_trn:conventional_rail")))
            return true;
        if (getEntitiesAhead(undefined, stoppingDistance, lookAhead, ["pod_trn_stock"]))
            return true;
    });
}
function findOffsetFromLine(turning, heading, position, slope) {
    const offset = { x: 0, y: 0, z: 0 };
    if (turning) {
        // TODO: Maybe in future
        // Logic for calculating position on arc
        return offset;
    }
    if (heading === "North" || heading === "South")
        offset.x = 0.5 - position.x;
    else if (heading === "West" || heading === "East")
        offset.z = 0.5 - position.z;
    else if (heading === "NorthWest" || heading === "SouthEast") {
        const point = (position.x + position.z) / 2;
        offset.x = point - position.x;
        offset.z = point - position.z;
    }
    else if (heading === "NorthEast" || heading === "SouthWest") {
        const point = (1 - position.x + position.z) / 2;
        offset.x = 1 - point - position.x;
        offset.z = point - position.z;
    }
    if (!slope) {
        offset.y = DESIRED_FLAT_Y_LEVEL - position.y;
    }
    if (slope === 1) {
        if (position.y < DESIRED_FLAT_Y_LEVEL)
            offset.y = DESIRED_FLAT_Y_LEVEL - position.y;
        if (position.y > DESIRED_SLOPE_1_Y_LEVEL_MAX)
            offset.y = DESIRED_SLOPE_1_Y_LEVEL_MAX - position.y;
    }
    return offset;
}
function getTrainHeading(e, directions, forwards = true) {
    if (!directions?.checked)
        return;
    const rot = e.getRotation();
    const traversables = directions.connectedTo
        .filter(d => directions[d] && directions[d].traversable)
        .map(d => ({
        direction: d,
        degrees: railDirectionToVector2(d), ...directions[d]
    }));
    // Determining heading of train
    // Heading is based off the closest cardinal or corner cardinal to the train's heading
    const currentHeading = yRotationToDirection(!forwards ? rot.y : (rot.y + 180));
    if (!currentHeading)
        return;
    let headingDetails = traversables.find(d => d.direction === currentHeading);
    let turning = false;
    if (!headingDetails) {
        headingDetails = traversables.find(d => directionRelativeTo("NorthWest", currentHeading) === d.direction);
        turning = "left";
    }
    if (!headingDetails) {
        headingDetails = traversables.find(d => directionRelativeTo("NorthEast", currentHeading) === d.direction);
        turning = "right";
    }
    return [headingDetails, turning];
}
function calculateEntityIncrement(e, directions, amount) {
    if (!directions?.checked)
        return;
    const forwards = amount >= 0;
    const trainRotation = e.getRotation();
    const heading = getTrainHeading(e, directions, forwards);
    if (!heading)
        return;
    const [headingDetails, turning] = heading;
    if (!headingDetails?.block)
        return;
    const slope = directions.details.type === "conventional" ? directions.details.slope : 0;
    if (slope) {
        const headingMatches = headingDetails.direction === directions.details.cardinal;
        trainRotation.x = headingMatches ? SLOPE_ANGLE : -SLOPE_ANGLE;
        trainRotation.x *= forwards ? 1 : -1;
    }
    else {
        trainRotation.x = 0;
    }
    // Don't process beyond this point if amount is zero
    if (!amount) {
        // To ensure stopped trains maintain angle between reloads
        e.setRotation(trainRotation);
        return;
    }
    const lookAhead = isHeadingAllowed(e, forwards, headingDetails.direction, headingDetails);
    if (!lookAhead) {
        // For locomotives this should cause throttle to decrease by 1 step each time heading is disallowed
        if (e.matches({ families: ["pod_trn_locomotive"] })) ;
        return;
    }
    const launchEntities = Math.abs(amount) > 4 * DELTA;
    const playerInFront = launchEntities ? getEntitiesAhead(undefined, getStoppingDistance(e.typeId), lookAhead, ["player"]) : [];
    const mobInFront = launchEntities ? getEntitiesAhead(undefined, getStoppingDistance(e.typeId), lookAhead, ["mob"]) : [];
    const [_whole, decimal] = splitCoordinates(e.location);
    const desiredAngle = headingDetails.degrees.y + (forwards ? 180 : 0); //+ (turning ? TURN_ANGLE * (turning === "left" ? -1 : 1) : 0)
    const TURN_CONSTANT = 4;
    if (trainRotation.y !== desiredAngle || turning) {
        trainRotation.y = interpolateDegrees(trainRotation.y, desiredAngle, Math.min(1, Math.abs(amount) * TURN_CONSTANT));
    }
    // Old immediate turn system
    // if (turning) {
    //     trainRotation.y += TURN_ANGLE * (turning === "left" ? -1 : 1)
    // }
    const forwardVelocity = calculateVelocity(amount, trainRotation);
    const alignmentVelocity = findOffsetFromLine(turning, headingDetails.direction, decimal, slope);
    const CORRECTION_CONSTANT = 1.0;
    alignmentVelocity.x *= CORRECTION_CONSTANT;
    alignmentVelocity.y *= CORRECTION_CONSTANT;
    alignmentVelocity.z *= CORRECTION_CONSTANT;
    return {
        entity: e,
        rotation: trainRotation,
        impulse: addVector(forwardVelocity, alignmentVelocity),
        entitiesHit: [...playerInFront || [], ...mobInFront || []]
    };
}
function getEntity(id) {
    try {
        if (!id)
            return null;
        const hitchedEntity = world.getEntity(id);
        if (!hitchedEntity)
            return null;
        if (!hitchedEntity?.isValid())
            return undefined;
        return hitchedEntity;
    }
    catch (event) {
        return undefined;
    }
}
function validateHitch(e) {
    const [to] = getHitchInfo(e);
    const distanceToHitchedTo = to?.isValid() ? getDistance(e.location, to.location) : 0;
    if (distanceToHitchedTo >= 6) {
        // log("Distance unhitch")
        unhitch(to, e);
        return 0;
    }
    let difference = 4 - distanceToHitchedTo;
    if (to?.getRotation().y !== e.getRotation().y)
        return 0;
    // TODO will need to be adjusted if other stopping distances need to be accommodated.
    return difference;
}
function getHitchInfo(e) {
    return [getEntity(e.getDynamicProperty("hitched_to")), getEntity(e.getDynamicProperty("hitched_by"))];
}
function getControllingRider(e) {
    const rideable = e.getComponent(EntityRideableComponent.componentId);
    return rideable.getRiders().at(rideable.controllingSeat);
}
function tryAddFuel(e, player) {
    const usesFuel = getUsesFuel(e.typeId);
    if (!usesFuel)
        return false;
    if (e.getProperty("pod_trn:fuel") === 1)
        return false;
    const equippable = player.getComponent(EntityEquippableComponent.componentId);
    if (!equippable.isValid())
        return false;
    const slot = equippable.getEquipmentSlot(EquipmentSlot.Mainhand);
    if (!slot.isValid() || !slot.hasItem())
        return false;
    const rememberTypeId = slot.typeId;
    const source = usesFuel.sources.find(d => d.typeId === rememberTypeId);
    const fills = source?.fills;
    if (!fills)
        return false;
    system.runTimeout(() => {
        if (!slot.isValid() || !slot.hasItem() || slot.typeId !== rememberTypeId)
            return;
        e.setProperty("pod_trn:fuel", Math.min(1, e.getProperty("pod_trn:fuel") + fills));
        e.dimension.playSound("random.pop", e.location, { pitch: 0.4 });
        if (player.getGameMode() === "creative")
            return;
        if (slot.amount === 1) {
            slot.setItem(source?.returnsBucket ? new ItemStack("minecraft:bucket") : undefined);
            return;
        }
        slot.amount--;
    });
    return true;
}
function interactLocomotive(e) {
    markPlayer(e.player);
    const rider = getControllingRider(e.target);
    // Override to allow custom handling for handcar (technically anything that doesn't have a horn)
    if (!getHorn(e.target.typeId)) {
        if (e.player.id === rider?.id) {
            system.runTimeout(() => {
                try {
                    e.target.dimension.playSound("random.lever_click", e.target.location, { pitch: !e.target.getDynamicProperty("invert_handcar") ? 0.5 : 0.6 });
                }
                catch (e) { }
            });
            e.target.setDynamicProperty("invert_handcar", !e.target.getDynamicProperty("invert_handcar"));
            e.cancel = true;
        }
        return;
    }
    if (!e.player.isSneaking && tryAddFuel(e.target, e.player)) {
        e.cancel = true;
        return;
    }
    if ((e.player.id === rider?.id) || (e.player.isSneaking && !rider)) {
        system.runTimeout(() => openLocomotiveConfig(e.target, e.player).catch());
        e.cancel = true;
    }
}
async function openLocomotiveConfig(e, p) {
    const controller = p ?? getControllingRider(e);
    if (controller?.typeId !== "minecraft:player")
        return;
    const result = await new ModalFormData()
        .title("form.locomotive.title")
        // .textField("form.locomotive.name", "", e.getDynamicProperty(Properties.Name) as string ?? "Player's Train")
        .slider({ rawtext: [{ translate: "form.locomotive.speed" }] }, THROTTLE_MIN, THROTTLE_MAX, 1, e.getDynamicProperty(Properties.Throttle) ?? 0)
        // .toggle("form.locomotive.horn", e.getDynamicProperty(Properties.Horn) as boolean ?? true)
        // .toggle("form.locomotive.respect_signals", e.getDynamicProperty(Properties.RespectSignals) as boolean ?? true)
        .toggle("form.locomotive.look_throttle", e.getDynamicProperty(Properties.LookThrottle) ?? false)
        .submitButton("form.locomotive.submit")
        .show(controller);
    if (result.canceled)
        return;
    const [/*name,*/ speed, /*autoHorn,*/ /*respectSignals,*/ lookThrottle] = result.formValues;
    setSettings(e, { speed, lookThrottle });
}
function unhitch(to, by) {
    by.setDynamicProperty("hitched_to", undefined);
    to.setDynamicProperty("hitched_by", undefined);
}
function hitch(to, by) {
    if (getEntity(by.getDynamicProperty("hitched_to")))
        return false;
    if (!(getCanHitch(by.typeId)) || !(getCanHitch(to.typeId)))
        return false;
    // Hacky fix to prevent overlapping too hard
    if (getDistance(to.location, by.location) < (getStoppingDistance(to.typeId) + getStoppingDistance(by.typeId)) / 2 - 1)
        return false;
    const yRotBy = ((by.getRotation().y % 360) + 360) % 360;
    const yRotTo = ((to.getRotation().y % 360) + 360) % 360;
    const diff = Math.abs(yRotBy - yRotTo);
    if (diff > 90 && diff < 270)
        return false;
    to.dimension.playSound("dig.chain", to.location, { pitch: 0.6 });
    by.setDynamicProperty("hitched_to", to.id);
    to.setDynamicProperty("hitched_by", by.id);
    return true;
}
function interactTrainCar(e) {
    markPlayer(e.player);
    if (e.target.typeId === "pod_trn:gondola_car" || e.target.typeId === "pod_trn:flatbed_car")
        handleStackContainerCarInteraction(e);
}
function handleStackContainerCarInteraction(e) {
    const equippable = e.player.getComponent(EntityEquippableComponent.componentId);
    if (!equippable.isValid())
        return;
    const item = equippable.getEquipment(EquipmentSlot.Mainhand);
    system.runTimeout(() => {
        if (item && !e.player.isSneaking) {
            const added = addItemToStackContainerCar(e.target, item);
            if (!added)
                return;
            e.player.dimension.playSound("random.pop", e.player.location);
            if (e.player.getGameMode() === "creative")
                return;
            if (item.amount === added) {
                equippable.setEquipment(EquipmentSlot.Mainhand, undefined);
                return;
            }
            item.amount -= added;
            equippable.setEquipment(EquipmentSlot.Mainhand, item);
            return;
        }
        const extractedItem = removeItemFromStackContainerCar(e.target);
        if (!extractedItem)
            return;
        const inventory = e.player.getComponent(EntityInventoryComponent.componentId);
        if (inventory?.isValid()) {
            const addItem = inventory.container?.addItem(extractedItem);
            if (addItem) {
                e.player.dimension.spawnItem(addItem, e.player.location);
            }
            e.player.dimension.playSound("random.pop", e.player.location);
            return;
        }
        e.player.dimension.spawnItem(extractedItem, e.player.location);
    });
}
function addItemToStackContainerCar(e, item) {
    if (!item)
        return false;
    switch (e.typeId) {
        case "pod_trn:gondola_car": return addItemToGondola(e, item);
        case "pod_trn:flatbed_car": return addItemToFlatbed(e, item);
    }
    throw new Error("Invalid entity");
}
function addItemToGondola(e, item) {
    const initialGondolaItem = GondolaItems[item.typeId];
    if (!initialGondolaItem)
        return false;
    const actualGondolaItem = initialGondolaItem.convert ? GondolaItems[initialGondolaItem.to] : initialGondolaItem;
    if (!actualGondolaItem || actualGondolaItem.convert)
        return false;
    const factor = initialGondolaItem.convert ? initialGondolaItem.amount : 1;
    const checkTypeId = initialGondolaItem.convert ? initialGondolaItem.to : item.typeId;
    const count = e.getDynamicProperty("count") ?? 0;
    const currentTypeId = count ? e.getDynamicProperty("type_id") : undefined;
    // Reject incompatible items, the gondola can only contain one type of item at a time
    if (count && currentTypeId !== checkTypeId)
        return false;
    const max = new ItemStack(checkTypeId).maxAmount * StackContainerCapacity;
    const freeSpace = max - count;
    const amountToAdd = Math.min(Math.floor(freeSpace / factor) * factor, item.amount * factor);
    const newCount = count + amountToAdd;
    e.setDynamicProperty("count", newCount);
    e.setDynamicProperty("type_id", checkTypeId);
    e.setProperty("pod_trn:cargo_type", actualGondolaItem.cargoType);
    e.setProperty("pod_trn:layers", newCount === 0 ? 0 : (Math.floor(newCount / max * 4) + 1));
    e.setProperty("pod_trn:red", actualGondolaItem.colour.r);
    e.setProperty("pod_trn:green", actualGondolaItem.colour.g);
    e.setProperty("pod_trn:blue", actualGondolaItem.colour.b);
    return amountToAdd / factor;
}
function addItemToFlatbed(e, item) {
    const flatbedItem = FlatbedItems[item.typeId];
    if (!flatbedItem)
        return false;
    const count = e.getDynamicProperty("count") ?? 0;
    const currentTypeId = count ? e.getDynamicProperty("type_id") : undefined;
    // Reject incompatible items, the gondola can only contain one type of item at a time
    if (count && currentTypeId !== item.typeId)
        return false;
    const max = new ItemStack(item.typeId).maxAmount * StackContainerCapacity;
    const freeSpace = max - count;
    const amountToAdd = Math.min(freeSpace, item.amount);
    const newCount = count + amountToAdd;
    e.setDynamicProperty("count", newCount);
    e.setDynamicProperty("type_id", item.typeId);
    e.setProperty("pod_trn:cargo_type", FlatbedCargoTypes.findIndex(d => d === flatbedItem.cargoType));
    e.setProperty("pod_trn:layers", newCount === 0 ? 0 : (Math.floor(newCount / max * 4) + 1));
    return amountToAdd;
}
function removeItemFromStackContainerCar(e) {
    const count = e.getDynamicProperty("count") ?? 0;
    const currentTypeId = count ? e.getDynamicProperty("type_id") : undefined;
    if (!currentTypeId)
        return false;
    const item = new ItemStack(currentTypeId);
    const max = item.maxAmount * StackContainerCapacity;
    const removedCount = Math.min(item.maxAmount, count);
    const newCount = count - removedCount;
    item.amount = removedCount;
    e.setDynamicProperty("count", newCount);
    e.setProperty("pod_trn:layers", newCount === 0 ? 0 : (Math.floor(newCount / max * 4) + 1));
    return item;
}
const playerMap = new Map;
function markPlayer(p) {
    playerMap.set(p.id, system.currentTick);
}
// Player clicking (long tapping) on trains and cars
world.beforeEvents.playerInteractWithEntity.subscribe(e => {
    if (e.target.matches({ families: ["pod_trn_stock"] }) || e.target.matches({ families: ["player"] })) {
        const equippable = e.player.getComponent(EntityEquippableComponent.componentId);
        if (!e.player.isSneaking && equippable?.isValid() && equippable.getEquipment(EquipmentSlot.Mainhand)?.typeId === "pod_trn:railroad_wrench") {
            if (e.target.typeId === "minecraft:player") {
                system.runTimeout(() => {
                    e.target.dimension.playSound("mob.goat.prepare_ram.screamer", e.target.location);
                });
                e.player.sendMessage({ translate: "pod_trn.easter_egg" });
                return;
            }
            const [to, by] = getHitchInfo(e.target);
            if (to === null && by === null) {
                const rot = e.target.getRotation();
                e.cancel = true;
                system.runTimeout(() => {
                    e.target.dimension.playSound("block.itemframe.add_item", e.target.location, { pitch: 0.3 });
                    e.target.setRotation({ x: rot.x, y: rot.y + 180 });
                });
                return;
            }
        }
    }
    if (e.target.matches({ families: ["pod_trn_locomotive"] }))
        return interactLocomotive(e);
    if (e.target.matches({ families: ["pod_trn_train_car"] }))
        return interactTrainCar(e);
});
// Punch train and get it to drop itself (also does decoupling)
world.afterEvents.entityHitEntity.subscribe(e => {
    if (!(e.damagingEntity instanceof Player))
        return false;
    const isCreative = e.damagingEntity.getGameMode() === GameMode.creative;
    if (e.hitEntity.matches({ families: ["pod_trn_stock"] })) {
        markPlayer(e.damagingEntity);
        // If already shaking trigger removal (drops in survival)
        if (e.hitEntity.hasComponent("minecraft:is_shaking"))
            return popTrain(e.hitEntity, isCreative);
        // Start shaking
        e.hitEntity.triggerEvent("pod_trn:start_shake");
        // Also trigger decouple here
        const [_to, by] = getHitchInfo(e.hitEntity);
        if (!by?.isValid())
            return;
        unhitch(e.hitEntity, by);
        e.hitEntity.dimension.playSound("dig.chain", e.hitEntity.location, { pitch: 0.8 });
        trainSetProcess(e.hitEntity, -20); // Cause the train to move backwards
    }
});
// Stop break through train
function playerRecentlyInteractedWithTrain(e) {
    if (playerMap.get(e.player.id) === system.currentTick)
        e.cancel = true;
}
world.beforeEvents.playerBreakBlock.subscribe(playerRecentlyInteractedWithTrain);
world.beforeEvents.playerInteractWithBlock.subscribe(playerRecentlyInteractedWithTrain);

const northSouth = { x: 0.5 };
const eastWest = { z: 0.5 };
const diagNorthWest = { z: 0.25, x: 0.25 };
const diagNorthEast = { z: 0.5, x: 0.5 };
function matchingContents(a, b) {
    return a.every(d => b.includes(d));
}
const trainPlacer = {
    onUseOn(e) {
        const railDirections = getRailDirections(e.block, e.usedOnBlockPermutation);
        if (!isPlacementSafe(e.block, railDirections, e.itemStack.typeId))
            return;
        // Place locomotive
        const locomotiveId = e.itemStack.typeId;
        const variant = e.itemStack.getDynamicProperty("variant") ?? getInitialVariant(e.itemStack.typeId);
        const facing = yRotationToDirection(e.source.getRotation().y);
        // Remove all other trains, testing line
        // e.source.dimension.getEntities({families: ["pod_trn_locomotive"]}).forEach(d => d.remove())
        let location;
        let rot;
        if (matchingContents(railDirections.connectedTo, ["North", "South"])) {
            location = addVector(e.block, {
                ...northSouth,
                y: 0.2
            });
            if (facing === "West" || facing === "North")
                rot = { x: 0, y: 0 };
            else
                rot = { x: 0, y: -180 };
        }
        else if (matchingContents(railDirections.connectedTo, ["West", "East"])) {
            location = addVector(e.block, {
                ...eastWest,
                y: 0.2
            });
            if (facing === "South" || facing === "West")
                rot = { x: 0, y: -90 };
            else
                rot = { x: 0, y: 90 };
        }
        else if (matchingContents(railDirections.connectedTo, ["NorthWest", "SouthEast"])) {
            location = addVector(e.block, {
                ...diagNorthWest,
                y: 0.2
            });
            if (facing === "SouthEast" || facing === "NorthEast")
                rot = { x: 0, y: 135 };
            else
                rot = { x: 0, y: -45 };
        }
        else if (matchingContents(railDirections.connectedTo, ["NorthEast", "SouthWest"])) {
            location = addVector(e.block, {
                ...diagNorthEast,
                y: 0.2
            });
            if (facing === "SouthWest" || facing === "NorthWest")
                rot = { x: 0, y: 45 };
            else
                rot = { x: 0, y: 225 };
        }
        if (location && rot) {
            e.block.dimension.playSound("block.lantern.fall", e.block, { pitch: 0.2 });
            e.block.dimension.runCommand(`/summon ${locomotiveId} ${location.x} ${location.y} ${location.z} ${rot.y} ${rot.x} pod_trn:${variants[variant]}`);
        }
        if (e.source.getGameMode() === "creative")
            return;
        const equipmentComponent = e.source.getComponent(EntityEquippableComponent.componentId);
        if (!equipmentComponent?.isValid())
            return;
        equipmentComponent.setEquipment(EquipmentSlot.Mainhand, new ItemStack("air"));
    }
};

/*
const buttonExample = {
  text: String | RawMessage -> {translate: "pod.example"},
  icon: String -> "textures/items/iron_ingot", //can be undefined, button will have no icon then.
  action: function with the player as parameter -> showPage("main", 0, []), //can be any function or undefined. If undefined, the button will close the menu.
};
*/
/**
 * Shows a page to a player and adds all specified buttons to it with their actions
 * @param player
 * @param title Title of the page
 * @param body Body of the page
 * @param buttons Buttons as array
 */
const showPage = async (player, title, body, buttons) => {
    if (!buttons || !buttons.length) {
        console.warn(`Failed to show Page: Requires Button.`);
        return;
    }
    const form = new ActionFormData();
    form.title(title);
    form.body(body);
    for (const button of buttons) {
        form.button(button.text, button.icon);
    }
    const response = await form.show(player);
    if (response.canceled || response.selection === undefined)
        return;
    const action = buttons[response.selection].action;
    if (!action)
        return;
    action(player);
};

world.afterEvents.playerSpawn.subscribe(e => {
    if (e.player.getDynamicProperty("given_guide_book")) {
        return;
    }
    e.player.setDynamicProperty("given_guide_book", true);
    e.player.dimension.spawnItem(new ItemStack("pod_trn:guide_book"), e.player.location);
});
const guide = {
    async onUse(e) {
        await showMainPage(e.source);
    }
};
const namespace = "pod_trn";
const defaultBackButton = {
    text: { translate: `${namespace}.guide.back_button` },
    action: (player) => showMainPage(player),
};
const railPageButton = {
    icon: "textures/podcrash/trains/icons/rails_icon.png",
    text: { translate: `${namespace}.guide.rails.title` },
    action: (player) => showGuidePage(player, "rails", railPageButtons),
};
const railBackPageButton = {
    ...railPageButton,
    icon: undefined,
    text: { translate: `${namespace}.guide.back_button` }
};
const railPageButtons = [
    {
        icon: "textures/podcrash/trains/icons/rail_icon.png",
        text: { translate: `${namespace}.guide.conventional_rails.title` },
        action: (player) => showGuidePage(player, "conventional_rails", [railBackPageButton]),
    },
    {
        icon: "textures/podcrash/trains/icons/embedded_rail_icon.png",
        text: { translate: `${namespace}.guide.embedded_rails.title` },
        action: (player) => showGuidePage(player, "embedded_rails", [railBackPageButton]),
    },
    {
        icon: "textures/podcrash/trains/icons/switch_icon.png",
        text: { translate: `${namespace}.guide.switch_rails.title` },
        action: (player) => showGuidePage(player, "switch_rails", [railBackPageButton]),
    },
    {
        icon: "textures/podcrash/trains/icons/buffer_icon.png",
        text: { translate: `${namespace}.guide.buffer_rails.title` },
        action: (player) => showGuidePage(player, "buffer_rails", [railBackPageButton]),
    },
    defaultBackButton
];
const rollingStockPageButton = {
    icon: "textures/podcrash/trains/icons/rolling_stock_icon.png",
    text: { translate: `${namespace}.guide.rolling_stock.title` },
    action: (player) => showGuidePage(player, "rolling_stock", rollingStockPageButtons),
};
const rollingStockBackPageButton = {
    ...rollingStockPageButton,
    icon: undefined,
    text: { translate: `${namespace}.guide.back_button` }
};
const operatingEnginesButton = {
    icon: "textures/podcrash/trains/icons/operating_engines_icon.png",
    text: { translate: `${namespace}.guide.operating_engines.title` },
    action: (player) => showGuidePage(player, "operating_engines", operatingEnginesButtons),
};
const operatingEnginesBackPageButton = {
    ...operatingEnginesButton,
    icon: undefined,
    text: { translate: `${namespace}.guide.back_button` }
};
const operatingEnginesButtons = [
    {
        icon: "textures/podcrash/trains/icons/coupling_icon.png",
        text: { translate: `${namespace}.guide.coupling.title` },
        action: (player) => showGuidePage(player, "coupling", [operatingEnginesBackPageButton]),
    },
    {
        icon: "textures/podcrash/trains/icons/train_horn_icon.png",
        text: { translate: `${namespace}.guide.train_horns.title` },
        action: (player) => showGuidePage(player, "train_horns", [operatingEnginesBackPageButton]),
    },
    {
        icon: "textures/podcrash/trains/items/diesel_bucket.png",
        text: { translate: `${namespace}.guide.fuel_bucket.title` },
        action: (player) => showGuidePage(player, "fuel_bucket", [operatingEnginesBackPageButton]),
    },
    {
        icon: "textures/podcrash/trains/items/railroad_wrench.png",
        text: { translate: `${namespace}.guide.railroad_wrench.title` },
        action: (player) => showGuidePage(player, "railroad_wrench", [operatingEnginesBackPageButton]),
    },
    rollingStockBackPageButton
];
const rollingStockPageButtons = [
    {
        icon: "textures/podcrash/trains/icons/locomotives.png",
        text: { translate: `${namespace}.guide.locomotives.title` },
        action: (player) => showGuidePage(player, "locomotives", locomotiveButtons),
    },
    {
        icon: "textures/podcrash/trains/icons/cars_icon.png",
        text: { translate: `${namespace}.guide.cars.title` },
        action: (player) => showGuidePage(player, "cars", trainCarButtons),
    },
    operatingEnginesButton,
    defaultBackButton
];
const locomotivePageBackButton = {
    text: { translate: `${namespace}.guide.back_button` },
    icon: undefined,
    action: (player) => showGuidePage(player, "locomotives", locomotiveButtons),
};
const trainCarPageBackButton = {
    text: { translate: `${namespace}.guide.back_button` },
    icon: undefined,
    action: (player) => showGuidePage(player, "cars", trainCarButtons),
};
const locomotiveButtons = [
    {
        icon: "textures/podcrash/trains/icons/steam_locomotive_icon.png",
        text: { translate: `${namespace}.guide.steam_locomotive.title` },
        action: (player) => showGuidePage(player, "steam_locomotive", [locomotivePageBackButton]),
    },
    {
        icon: "textures/podcrash/trains/icons/deisel_locomotive_icon.png",
        text: { translate: `${namespace}.guide.diesel_locomotive.title` },
        action: (player) => showGuidePage(player, "diesel_locomotive", [locomotivePageBackButton]),
    },
    {
        icon: "textures/podcrash/trains/icons/handcar_icon.png",
        text: { translate: `${namespace}.guide.handcar.title` },
        action: (player) => showGuidePage(player, "handcar", [rollingStockBackPageButton]),
    },
    rollingStockBackPageButton
];
const trainCarButtons = [
    {
        icon: "textures/podcrash/trains/icons/flat_car_icon.png",
        text: { translate: `${namespace}.guide.flatbed_car.title` },
        action: (player) => showGuidePage(player, "flatbed_car", [trainCarPageBackButton]),
    },
    {
        icon: "textures/podcrash/trains/icons/gondola_car_icon.png",
        text: { translate: `${namespace}.guide.gondola_car.title` },
        action: (player) => showGuidePage(player, "gondola_car", [trainCarPageBackButton]),
    },
    {
        icon: "textures/podcrash/trains/icons/animal_car_icon.png",
        text: { translate: `${namespace}.guide.animal_car.title` },
        action: (player) => showGuidePage(player, "animal_car", [trainCarPageBackButton]),
    },
    {
        icon: "textures/podcrash/trains/icons/passenger_car_icon.png",
        text: { translate: `${namespace}.guide.passenger_car.title` },
        action: (player) => showGuidePage(player, "passenger_car", [trainCarPageBackButton]),
    },
    {
        icon: "textures/podcrash/trains/icons/caboose_icon.png",
        text: { translate: `${namespace}.guide.caboose.title` },
        action: (player) => showGuidePage(player, "caboose", [trainCarPageBackButton]),
    },
    rollingStockBackPageButton
];
const otherPageButton = {
    icon: "textures/podcrash/trains/icons/other_blocks_icon.png",
    text: { translate: `${namespace}.guide.other.title` },
    action: (player) => showGuidePage(player, "other", otherPageButtons),
};
const otherBackPageButton = {
    ...otherPageButton,
    icon: undefined,
    text: { translate: `${namespace}.guide.back_button` }
};
const otherPageButtons = [
    {
        icon: "textures/podcrash/trains/icons/trestle_icon.png",
        text: { translate: `${namespace}.guide.trestles.title` },
        action: (player) => showGuidePage(player, "trestles", [otherBackPageButton]),
    },
    {
        icon: "textures/podcrash/trains/icons/platform_icon.png",
        text: { translate: `${namespace}.guide.platforms.title` },
        action: (player) => showGuidePage(player, "platforms", [otherBackPageButton]),
    },
    {
        icon: "textures/podcrash/trains/icons/fuel_tank_icon.png",
        text: { translate: `${namespace}.guide.fuel_tank.title` },
        action: (player) => showGuidePage(player, "fuel_tank", [otherBackPageButton]),
    },
    {
        icon: "textures/podcrash/trains/icons/boiler_icon.png",
        text: { translate: `${namespace}.guide.boiler.title` },
        action: (player) => showGuidePage(player, "boiler", [otherBackPageButton]),
    },
    {
        icon: "textures/podcrash/trains/icons/cab_frame_icon.png",
        text: { translate: `${namespace}.guide.cab_frame.title` },
        action: (player) => showGuidePage(player, "cab_frame", [otherBackPageButton]),
    },
    {
        icon: "textures/podcrash/trains/icons/truck_icon.png",
        text: { translate: `${namespace}.guide.truck.title` },
        action: (player) => showGuidePage(player, "truck", [otherBackPageButton]),
    },
    defaultBackButton
];
const functionalBlocksPageButton = {
    icon: "textures/podcrash/trains/icons/functional_blocks_icon.png",
    text: { translate: `${namespace}.guide.functional_blocks.title` },
    action: (player) => showGuidePage(player, "functional_blocks", functionalBlocksPageButtons),
};
const functionalBlocksBackPageButton = {
    ...functionalBlocksPageButton,
    icon: undefined,
    text: { translate: `${namespace}.guide.back_button` }
};
const functionalBlocksPageButtons = [
    {
        icon: "textures/podcrash/trains/icons/fuel_distillery_icon.png",
        text: { translate: `${namespace}.guide.fuel_distillery.title` },
        action: (player) => showGuidePage(player, "fuel_distillery", [functionalBlocksBackPageButton]),
    },
    {
        icon: "textures/podcrash/trains/icons/station_marker_icon.png",
        text: { translate: `${namespace}.guide.station_markers.title` },
        action: (player) => showGuidePage(player, "station_markers", [functionalBlocksBackPageButton]),
    },
    {
        icon: "textures/podcrash/trains/icons/crossing_signal_icon.png",
        text: { translate: `${namespace}.guide.crossing_signals.title` },
        action: (player) => showGuidePage(player, "crossing_signals", [functionalBlocksBackPageButton]),
    },
    {
        icon: "textures/podcrash/trains/icons/train_signal_icon.png",
        text: { translate: `${namespace}.guide.train_signals.title` },
        action: (player) => showGuidePage(player, "train_signals", [functionalBlocksBackPageButton]),
    },
    {
        icon: "textures/podcrash/trains/icons/signal_box_icon.png",
        text: { translate: `${namespace}.guide.signal_boxes.title` },
        action: (player) => showGuidePage(player, "signal_boxes", [functionalBlocksBackPageButton]),
    },
    defaultBackButton
];
const mainPageButtons = [
    railPageButton,
    rollingStockPageButton,
    functionalBlocksPageButton,
    otherPageButton,
    {
        text: { translate: `${namespace}.guide.exit_button` },
    },
];
async function showMainPage(player) {
    await showGuidePage(player, "main", mainPageButtons);
}
async function showGuidePage(player, id, buttons) {
    await showPage(player, { translate: `${namespace}.guide.${id}.title` }, { translate: `${namespace}.guide.${id}.body`, with: [`\n`] }, buttons);
}

const stationMarker = {
    onPlayerInteract(e) {
        const value = e.block.permutation.getState("pod_trn:delay");
        e.block.setPermutation(e.block.permutation.withState("pod_trn:delay", value === 4 ? 0 : value + 1));
        e.dimension.playSound(value === 4 ? "block.copper_bulb.turn_off" : "block.copper_bulb.turn_on", e.block);
    },
    onTick(e) {
        e.block.setPermutation(e.block.permutation.withState("pod_trn:powered", !!e.block.getRedstonePower()));
    }
};

const NON_PRIMARY_DISTANCE = 3.5;
const VERTICAL_DISTANCE = 4;
function getComparison(cardinal, block) {
    switch (cardinal) {
        case Direction.South:
            return ({ z, x }) => z < block.z && x > block.x && x - block.x <= NON_PRIMARY_DISTANCE;
        case Direction.West:
            return ({ x, z }) => x > block.x && z > block.z && z - block.z <= NON_PRIMARY_DISTANCE;
        case Direction.East:
            return ({ x, z }) => x < block.x && z < block.z && block.z - z <= NON_PRIMARY_DISTANCE;
        default:
            return ({ z, x }) => z > block.z && x < block.x && block.x - x <= NON_PRIMARY_DISTANCE;
    }
}
function signalLocomotives(block, go) {
    const cardinal = getCardinal(block.permutation);
    const blockFacingY = mod(railDirectionToVector2(cardinal).y, 360);
    const comparison = getComparison(cardinal, block);
    const entities = block.dimension
        .getEntities({
        location: block.location,
        maxDistance: 15,
        families: ["pod_trn_signalable"]
    })
        .filter(e => {
        if (!e.getProperty("pod_trn:fuel"))
            return false;
        if (block.y < e.location.y - 1 || block.y - e.location.y > VERTICAL_DISTANCE)
            return false;
        const settings = getSettings(e);
        const throttle = settings[Properties.Throttle];
        const throttleWas = settings[Properties.ThrottleWas];
        const forwards = throttle !== 0 ? throttle > 0 : throttleWas >= 0;
        const trainFacingY = mod(forwards ? e.getRotation().y + 180 : e.getRotation().y, 360);
        return comparison(e.location) && trainFacingY === blockFacingY;
    });
    // Start branch
    if (go)
        for (const entity of entities) {
            const settings = getSettings(entity);
            if (settings[Properties.Throttle] !== 0 || !settings[Properties.ThrottleWas])
                continue;
            setSettings(entity, { speed: settings[Properties.ThrottleWas] });
        }
    // Stop branch
    else
        for (const entity of entities) {
            const settings = getSettings(entity);
            if (settings[Properties.Throttle] === 0)
                continue;
            setSettings(entity, { speed: 0 }, -1);
        }
}
const trainSignal = {
    onTick(e) {
        // Signal handling
        signalLocomotives(e.block, !!e.block.permutation.getState("pod_trn:go"));
    }
};

const MAX_CHECK_LIMIT = 16;
function processSignalBox(block, active) {
    let checkBlock = null;
    for (let i = 1; i < MAX_CHECK_LIMIT; i++) {
        checkBlock = block.above(i);
        if (!checkBlock?.isValid() || checkBlock.isAir)
            return;
        if (checkBlock.hasTag("pod_trn:train_signal")) {
            checkBlock.setPermutation(checkBlock.permutation.withState("pod_trn:go", active));
            checkBlock.dimension.playSound(active ? "tile.piston.in" : "tile.piston.out", checkBlock, { pitch: 2, volume: 0.5 });
        }
        else if (checkBlock.hasTag("pod_trn:crossing_signal")) {
            checkBlock.setPermutation(checkBlock.permutation.withState("pod_trn:muted", active));
        }
        else if (checkBlock.hasTag("pod_trn:switched_rail")) {
            toggle(checkBlock, active);
        }
    }
}
function toggle(block, force) {
    const active = block.permutation.getState("pod_trn:switched");
    if (force !== undefined && force === active)
        return;
    block.dimension.playSound("random.lever_click", block, {
        pitch: active ? 0.5 : 0.6,
        volume: force === undefined ? 1 : 0.2
    });
    block.setPermutation(block.permutation.withState("pod_trn:switched", !active));
    if (block.hasTag("pod_trn:signal_box"))
        processSignalBox(block, !active);
}
const switchComponent = {
    onPlayerInteract(e) {
        toggle(e.block);
    },
    onTick(e) {
        const power = e.block.getRedstonePower();
        if (power === undefined)
            return;
        if (e.block.permutation.getState("pod_trn:powered") === !!power)
            return;
        e.block.setPermutation(e.block.permutation.withState("pod_trn:powered", !!power));
        toggle(e.block, !!power);
    }
};

// Percentage chance of an item increasing the level
const items = {
    "minecraft:coal": 0.05,
    "minecraft:charcoal": 0.05,
    "minecraft:coal_block": 0.45,
    "minecraft:bone": 0.10,
    "minecraft:bone_meal": 0.04,
    "minecraft:bone_block": 0.36,
    "minecraft:dried_kelp": 0.02,
    "minecraft:dried_kelp_block": 0.18,
};
const MAX_LEVEL = 7;
const fuelDistillery = {
    onPlayerInteract(e) {
        const currentLevel = e.block.permutation.getState("pod_trn:level");
        const equippable = e.player?.getComponent(EntityEquippableComponent.componentId);
        if (!equippable?.isValid())
            return;
        const item = equippable.getEquipment(EquipmentSlot.Mainhand);
        if (!item)
            return;
        // Player extracts fuel into bucket from full distillery
        if (currentLevel === MAX_LEVEL) {
            if (item.typeId === "minecraft:bucket") {
                if (item.amount === 1) {
                    equippable.setEquipment(EquipmentSlot.Mainhand, new ItemStack("pod_trn:fuel_bucket"));
                }
                else {
                    item.amount--;
                    equippable.setEquipment(EquipmentSlot.Mainhand, item);
                    const fuelBucket = new ItemStack("pod_trn:fuel_bucket");
                    const inventory = e.player.getComponent(EntityInventoryComponent.componentId);
                    if (!inventory?.isValid() || inventory.container?.addItem(fuelBucket)) {
                        e.dimension.spawnItem(fuelBucket, e.player.location);
                    }
                }
                e.block.setPermutation(e.block.permutation.withState("pod_trn:level", 0));
                e.dimension.playSound("bucket.fill_lava", e.block, { pitch: 0.6 });
            }
            return;
        }
        const chance = items[item.typeId];
        if (!chance)
            return;
        if (e.player.getGameMode() !== "creative") {
            if (item.amount > 1) {
                item.amount--;
                equippable.setEquipment(EquipmentSlot.Mainhand, item);
            }
            else {
                equippable.setEquipment(EquipmentSlot.Mainhand, undefined);
            }
        }
        e.dimension.playSound("block.muddy_mangrove_roots.step", e.block, { pitch: 0.75 });
        if (Math.random() > chance)
            return;
        e.block.setPermutation(e.block.permutation.withState("pod_trn:level", currentLevel + 1));
    }
};

function worldInitialize(e) {
    e.blockComponentRegistry.registerCustomComponent("pod_trn:rail", rail);
    e.blockComponentRegistry.registerCustomComponent("pod_trn:switch", switchComponent);
    e.blockComponentRegistry.registerCustomComponent("pod_trn:crossing_signal", crossingSignal);
    e.blockComponentRegistry.registerCustomComponent("pod_trn:crossing_signal_top", crossingSignalTop);
    e.blockComponentRegistry.registerCustomComponent("pod_trn:train_signal", trainSignal);
    e.blockComponentRegistry.registerCustomComponent("pod_trn:trestle", trestle);
    e.blockComponentRegistry.registerCustomComponent("pod_trn:platform", platform);
    e.blockComponentRegistry.registerCustomComponent("pod_trn:station_marker", stationMarker);
    e.blockComponentRegistry.registerCustomComponent("pod_trn:fuel_distillery", fuelDistillery);
    e.itemComponentRegistry.registerCustomComponent("pod_trn:train_placer", trainPlacer);
    e.itemComponentRegistry.registerCustomComponent("pod_trn:guide_book", guide);
}
world.beforeEvents.worldInitialize.subscribe(worldInitialize);

const overworld = world.getDimension("overworld");
/**
 * Prefix used for the debugging tag (e.g. podcrash.Debug)
 */
const NAMESPACE = "podcrash";
/**
 * Parses a JSON Object to a formatted string
 * @param object
 * @returns object parsed to string
 */
function processObject(object) {
    if (!Object.keys(object).length)
        return "§7§oEmpty Object§r";
    let output = "{ ";
    for (const [key, value] of Object.entries(object)) {
        output += `§6${key}§7: §r${process(value)} `;
    }
    return `${output}}`;
}
function process(arg) {
    let output = arg?.toString();
    if (Array.isArray(arg))
        output = `[${arg.map((d) => process(d)).join(", ")}]`;
    else if (arg === null)
        output = `§6null`;
    else
        switch (typeof arg) {
            case "object":
                output = processObject(arg);
                break;
            case "string":
                output = `${arg.startsWith("Error") ? "§c" : '§a"§2'}${arg}§a${arg.startsWith("Error") ? "" : '"'}`;
                break;
            case "number":
                output = isNaN(arg) ? `§5§oNaN` : `§9${arg}`;
                break;
            case "boolean":
                output = `§6${output}`;
                break;
            case "undefined":
                output = `§6undefined`;
                break;
        }
    return `${output}§r`;
}
/**
 * Formats arguments with colour and outputs them to Minecraft chat via tellraw.
 *
 * @param args
 */
function log(...args) {
    overworld
        .runCommandAsync(`tellraw @a[tag=${NAMESPACE}.Debug] {"rawtext":[{"text":"${args
        .map((d) => process(d))
        .join(" ")
        .replaceAll('"', '\\"')}"}]}`)
        .then()
        .catch();
}

const dimensions = [
    world.getDimension("minecraft:overworld"),
    world.getDimension("minecraft:nether"),
    world.getDimension("minecraft:the_end")
];
async function trainLoop() {
    // Ensure compatible with all dimensions
    for (const dimension of dimensions) {
        dimension
            .getEntities({ families: ["pod_trn_locomotive"] })
            .forEach(e => {
            try {
                return locomotiveProcess(e);
            }
            catch (e) {
                log(`Error with locomotiveProcess: ${e}`);
            }
        });
        dimension
            .getEntities({ families: ["pod_trn_train_car"] })
            .forEach(e => {
            try {
                return carProcess(e);
            }
            catch (e) {
                log(`Error with carProcess: ${e}`);
            }
        });
    }
}
async function loop() {
    for (const player of world.getAllPlayers()) {
        const component = player.getComponent(EntityEquippableComponent.componentId);
        if (!component)
            continue;
        if (component.getEquipment(EquipmentSlot.Mainhand)?.typeId === "pod_trn:crossing_signal_top")
            component.setEquipment(EquipmentSlot.Mainhand, new ItemStack("pod_trn:oak_crossing_signal"));
    }
}
system.runInterval(() => trainLoop(), 1);
system.runInterval(() => loop(), 20);
